/*
##############################################################################
#
# MODULE:       r.randomwalk.c
#
# AUTHOR:       Martin Mergili
# CONTRIBUTORS: Massimiliano Alvioli, Ivan Marchesini,
#               Markus Metz, Matthias Rauter
#
# PURPOSE:      Conceptual tool for mass movement propagation
#               Routing script
#
# COPYRIGHT:    (c) 2010 - 2018 by the author
#               (c) 2010 - 2018 by the BOKU University, Vienna
#               (c) 2015 - 2018 by the University of Vienna
#               (c) 2000 - 2018 by the GRASS Development Team
#
# VERSION:      20180426 (26 April 2018)
#
#               This program is free software under the GNU General Public
#               License (>=v2). Read the file COPYING that comes with GRASS
#               for details.
#
##############################################################################
*/


// USE OF GRASS GIS ---------------------------------------------------------------------------

#define WITHGRASS


// LIBRARIES AND STRUCTURES -------------------------------------------------------------------

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <fcntl.h>
#include <time.h>
#include <string.h>

#ifdef WITHGRASS

    #include <grass/gis.h>
    #include <grass/gprojects.h>
    #include <grass/glocale.h>
    #include <grass/raster.h>
    #include <grass/segment.h>

#endif

struct ico { // input constants and control variables

    char *MAINMAPSET; int M; int N; double UDM; int UDP; double CSZ; double BDWEST; double BDNORTH;

    #ifdef WITHGRASS

        int NSEGCR; int NSEGS;

    #else

        double BDSOUTH;

    #endif

};


// DEFINING FUNCTIONS -------------------------------------------------------------------------


int **alloc_imatrix(int mrows, int mcols) { // function for allocating 2D integer arrays
    int **mm;
    int mi;
    mm = (int **) calloc(mrows, sizeof(int *));
    mm[0] = (int *) calloc(mrows * mcols, sizeof(int));
    for (mi = 1; mi < mrows; mi++)
    mm[mi] = mm[mi - 1] + mcols;
    return mm;
}

int fiparam ( FILE *gfparam ) { // function for reading integer parameters

    int gparam;
    char gsparam[1000], **gstrtodhlp=NULL;
    fgets(gsparam, 1000, gfparam);
    gparam=strtod(gsparam, gstrtodhlp);
    return gparam;
}

double fdparam ( FILE *gfparam ) { // function for reading double parameters

    double gparam;
    char gsparam[1000], **gstrtodhlp=NULL;
    fgets(gsparam, 1000, gfparam);
    gparam=strtod(gsparam, gstrtodhlp);
    return gparam;
}

char *fcparam ( FILE *gfparam ) { // function for reading character parameters

    char *gsparam = (char*) calloc ( 1000, sizeof(char));
    fgets(gsparam, 1000, gfparam);
    gsparam[strnlen(gsparam, 1010) - 1] = '\0';
    return gsparam;
}

double falpha ( double gbetax, double gbetay ) { // function for flow direction

    //int gfdir;
    double galpha;
    double GPI = 3.14159;

    if ( gbetax == 0 && gbetay == 0 ) galpha = 0;
    else if ( gbetax == 0 && gbetay > 0 ) galpha = 90;
    else if ( gbetax == 0 && gbetay < 0 ) galpha = 270;
    else if ( gbetax >= 0 ) galpha = atan( tan( gbetay ) / tan( gbetax )) * 180 / GPI;
    else if ( gbetax < 0 ) galpha = atan( tan( gbetay ) / tan( gbetax )) * 180 / GPI + 180;

    if ( galpha < 0 ) galpha += 360;
    if ( galpha > 360 ) galpha -= 360;

    /*if      (                   galpha <=  22.5 ) gfdir = 6;
    else if ( galpha >  22.5 && galpha <=  67.5 ) gfdir = 7;
    else if ( galpha >  67.5 && galpha <= 112.5 ) gfdir = 8;
    else if ( galpha > 112.5 && galpha <= 157.5 ) gfdir = 1;
    else if ( galpha > 157.5 && galpha <= 202.5 ) gfdir = 2;
    else if ( galpha > 202.5 && galpha <= 247.5 ) gfdir = 3;
    else if ( galpha > 247.5 && galpha <= 292.5 ) gfdir = 4;
    else if ( galpha > 292.5 && galpha <= 337.5 ) gfdir = 5;
    else if ( galpha > 337.5                    ) gfdir = 6;*/

    return galpha;
    //return gfdir;
}

#ifdef WITHGRASS

    SEGMENT finrasti ( char gm[1000], struct ico sico ) { // input of integer GRASS raster maps

        int gf, gx, gy, gin;
        CELL *gc;
        SEGMENT gseg_in;

        Segment_open (&gseg_in, G_tempfile(), sico.M, sico.N, sico.NSEGCR, sico.NSEGCR, sizeof(int), sico.NSEGS);

        gc = Rast_allocate_c_buf();
        gf = Rast_open_old ( gm, sico.MAINMAPSET );

        for ( gx = 0; gx < sico.M; gx++ ) {

            Rast_get_c_row ( gf, gc, gx );

            for ( gy = 0; gy < sico.N; gy++ ) {

                if ( !Rast_is_c_null_value ( gc + gy ) ) gin = ( int ) gc[gy];
                else gin = (int)sico.UDM;
                Segment_put(&gseg_in, &gin, gx, gy);
	    }
        }
        Segment_flush(&gseg_in);
        G_free ( gc );
        Rast_close ( gf );

        return gseg_in;
    }

    SEGMENT finrastd ( char gm[1000], struct ico sico ) { // input of double GRASS raster maps

        int gf, gx, gy;
        double gin;
        DCELL *gc;
        SEGMENT gseg_in;

        Segment_open (&gseg_in, G_tempfile(), sico.M, sico.N, sico.NSEGCR, sico.NSEGCR, sizeof(double), sico.NSEGS);

        gc = Rast_allocate_d_buf();
        gf = Rast_open_old ( gm, sico.MAINMAPSET );

        for ( gx = 0; gx < sico.M; gx++ ) {

            Rast_get_d_row ( gf, gc, gx );

            for ( gy = 0; gy < sico.N; gy++ ) {

                if ( !Rast_is_d_null_value ( gc + gy ) ) gin = ( double ) gc[gy];
                else gin = sico.UDM;
                Segment_put(&gseg_in, &gin, gx, gy);
	    }
        }

        Segment_flush(&gseg_in);
        G_free ( gc );
        Rast_close ( gf );

        return gseg_in;
    }

    void foutrasti ( char gm[1000], SEGMENT gseg_v, struct ico sico ) { // output of integer GRASS raster maps

        int gf, gx, gy;
        int gv;
        CELL *gc;

        if ( gm != NULL ) {
            gc = Rast_allocate_c_buf();
	    gf = Rast_open_fp_new( gm );
        }

        for ( gx = 0; gx < sico.M; gx++ ) {

            if ( gm != NULL ) {
	        for ( gy = 0; gy < sico.N; gy++ ) {
                    Segment_get(&gseg_v, &gv, gx, gy);
                    gc[gy] = (CELL) gv;
                }
	        Rast_put_c_row( gf, gc );
            }
        }
        G_free( gc );
        Rast_close( gf );

        return;
    }

    void foutrastd ( char gm[1000], SEGMENT gseg_v, struct ico sico ) { // output of double GRASS raster maps

        int gf, gx, gy;
        double gv;
        DCELL *gc;

        if ( gm != NULL ) {
            gc = Rast_allocate_d_buf();
	    gf = Rast_open_fp_new( gm );
        }

        for ( gx = 0; gx < sico.M; gx++ ) {

            if ( gm != NULL ) {
	        for ( gy = 0; gy < sico.N; gy++ ) {
                    Segment_get(&gseg_v, &gv, gx, gy);
                    gc[gy] = (DCELL) gv;
                }
	        Rast_put_d_row( gf, gc );
            }
        }
        G_free( gc );
        Rast_close( gf );

        return;
    }

#else

    double **alloc_dmatrix(int mrows, int mcols) { // function for allocating 2D double arrays
        double **mm;
        int mi;
        mm = (double **) calloc(mrows, sizeof(double *));
        mm[0] = (double *) calloc(mrows * mcols, sizeof(double));
        for (mi = 1; mi < mrows; mi++)
        mm[mi] = mm[mi - 1] + mcols;
        return mm;
    }

    int ***alloc_imatrix3 (int mrows, int mcols, int mdepths) { // function for allocating 3D double arrays
        int ***mm;
        int mi, mj;
        mm = (int***)calloc(mrows, sizeof(int**));

        for (mi=0; mi<mrows; mi++) {
            mm[mi] = (int**)calloc(mcols, sizeof(int*));

            for (mj=0; mj<mcols; mj++) {
                mm[mi][mj] = (int*)calloc(mdepths, sizeof(int));
            }
        }
        return mm;
    }

    void free_imatrix3(int ***mm,int mi,int mj)
    {
        int i,j;

        for(i=0;i<mi;i++)
        {
            for(j=0;j<mj;j++)
            {
                free(mm[i][j]);
            }
            free(mm[i]);
        }
        free(mm);
    }

    char *flparam ( FILE *gfparam ) { // function for reading ascii raster line

        int asclinemax = 100000;
        char *gsparam2 = (char*) calloc (( asclinemax + 20 ), sizeof(char));
        fgets(gsparam2, asclinemax, gfparam);
        gsparam2[strnlen(gsparam2, asclinemax+10) - 1] = '\0';
        return gsparam2;
    }

    double *finaschdr ( char *gname, struct ico sico ) { // function for input of ascii raster map header

        FILE *gfascii;
        char gpath[200], *tok;
        char delim[] = " \t\n";
        int gi;
        double *gparam;
        char *gtest, *gtest2;
        char **strtodhlp=NULL;

        gparam = (double*) calloc( 6, sizeof(double));

        sprintf(gpath, "%s", gname ); // writing name of ascii file to string
        gfascii=fopen(gpath, "r"); // opening ascii file

        for ( gi=0; gi<6; gi++ ) { // reading header lines and writing to array:
            gtest = fcparam ( gfascii );
            tok=strtok(gtest, delim);
            while ( tok != NULL ) {
                gtest2=tok;
                tok=strtok(NULL, delim);
            }
            gparam[gi] = strtod (gtest2, strtodhlp);
            free(gtest);
        }

        fclose( gfascii );
        return gparam;
    }

    int **finascvali ( char *gname, struct ico sico ) { // function for input of integer ascii raster map values

        FILE *gfascii;
        char *gpath, *tok;
        char delim[] = " \t\n";
        int gi, gx, gy, gtest3, **gparam;
        char **strtodhlp=NULL;

        char *gtest; // allocating memory to arrays
        char *gtest2;

        gpath = (char*) calloc ( 1000, sizeof(char)); // allocating memory to arrays:
        gparam = alloc_imatrix( sico.M, sico.N );

        sprintf(gpath, "%s", gname ); // writing name of ascii file to string
        gfascii=fopen(gpath, "r"); // opening ascii file
        for ( gi=0; gi<6; gi++ ) {
            gtest = fcparam ( gfascii ); // reading header lines
            free(gtest);
        }

        for ( gx=0; gx<sico.M; gx+=1 ) { // writing raster values of all rows and columns to array:

            gtest = flparam ( gfascii );
            tok = strtok(gtest, delim);
            for ( gy=0; gy<sico.N; gy+=1 ) {

                gtest2 = tok;
                tok = strtok(NULL, delim);
                gtest3 = strtod (gtest2, strtodhlp);
                gparam[gx][gy] = gtest3;
            }
            free(gtest);
        }
        fclose( gfascii );
        free(gpath);
        return gparam;
    }

    double **finascvald ( char *gname, struct ico sico ) { // function for input of double ascii raster map values

        FILE *gfascii;
        char *gpath, *tok;
        char delim[] = " \t\n";
        int gi, gx, gy, gtest3;
        double **gparam;
        char **strtodhlp=NULL;

        char *gtest; // allocating memory to arrays
        char *gtest2;

        gpath = (char*) calloc ( 1000, sizeof(char)); // allocating memory to arrays:
        gparam = alloc_dmatrix( sico.M, sico.N );

        sprintf(gpath, "%s", gname ); // writing name of ascii file to string
        gfascii=fopen(gpath, "r"); // opening ascii file
        for ( gi=0; gi<6; gi++ ) {
            gtest = fcparam ( gfascii ); // reading header lines
            free(gtest);
        }

        for ( gx=0; gx<sico.M; gx+=1 ) { // writing raster values of all rows and columns to array:

            gtest = flparam ( gfascii );
            tok = strtok(gtest, delim);
            for ( gy=0; gy<sico.N; gy+=1 ) {

                gtest2 = tok;
                tok = strtok(NULL, delim);
                gtest3 = strtod (gtest2, strtodhlp);
                gparam[gx][gy] = gtest3;
            }
            free(gtest);
        }
        fclose( gfascii );
        free(gpath);
        return gparam;
    }

    void foutasci ( int **gparam, char *goutmaps, char gname[1000], struct ico sico ) { // function for output of integer ascii raster maps

        FILE *gfascii;
        char gpath[1000];
        int gx, gy;

        sprintf(gpath, "%s%s.asc", goutmaps, gname ); // writing name of ascii file to string
        gfascii=fopen(gpath, "w"); // opening ascii file

        fprintf( gfascii, "ncols %i\n", sico.N ); // writing header to ascii file
        fprintf( gfascii, "nrows %i\n", sico.M );
        fprintf( gfascii, "xllcorner %.2f\n", sico.BDWEST );
        fprintf( gfascii, "yllcorner %.2f\n", sico.BDSOUTH );
        fprintf( gfascii, "cellsize %.2f\n", sico.CSZ );
        fprintf( gfascii, "NODATA_value %.2f\n", sico.UDM );

        for ( gx=0; gx<sico.M; gx+=1 ) { for ( gy=0; gy<sico.N; gy+=1 ) { // loop over all pixels:

            fprintf(gfascii, "%i ", gparam[gx][gy] ); // writing output to ascii file

        } fprintf(gfascii, "\n"); }
        fclose(gfascii);

        return;
    }

    void foutascd ( double **gparam, char *goutmaps, char gname[1000], struct ico sico ) { // function for output of double ascii raster maps

        FILE *gfascii;
        char gpath[1000];
        int gx, gy;

        sprintf(gpath, "%s%s.asc", goutmaps, gname ); // writing name of ascii file to string
        gfascii=fopen(gpath, "w"); // opening ascii file

        fprintf( gfascii, "ncols %i\n", sico.N ); // writing header to ascii file
        fprintf( gfascii, "nrows %i\n", sico.M );
        fprintf( gfascii, "xllcorner %.2f\n", sico.BDWEST );
        fprintf( gfascii, "yllcorner %.2f\n", sico.BDSOUTH );
        fprintf( gfascii, "cellsize %.2f\n", sico.CSZ );
        fprintf( gfascii, "NODATA_value %.2f\n", sico.UDM );

        for ( gx=0; gx<sico.M; gx+=1 ) { for ( gy=0; gy<sico.N; gy+=1 ) { // loop over all pixels:

            fprintf(gfascii, "%.2f ", gparam[gx][gy] ); // writing output to ascii file

        } fprintf(gfascii, "\n"); }
        fclose(gfascii);

        return;
}

#endif


// CALLING MAIN FUNCTION ----------------------------------------------------------------------


int main(int argc, char *argv[]) {


// DECLARING VARIABLES ------------------------------------------------------------------------


    char *PAR_PREFIX, *PAR_FILEPATH, *PAR_MAPPATH, *PAR_AOI, *PAR_ELEV, *PAR_FLOWDIR, *PAR_RELEASEMAP, *PAR_IMPOBJ, *PAR_IMPAREA,
        *PAR_SCOREMAP, *PAR_PROBMAP, path[1000], pathc[1000], pathd[1000], paths[1000], patho[1000], wkdir[1000];

    char *yint = (char*) calloc( 1000, sizeof(char));

    int PAR_AFLAG, PAR_MULT, PAR_BACK, PAR_CONN, PAR_SWALK, PAR_CTRLFLOWDIR, PAR_CTRLRELEASEMAP, PAR_CTRLPROBMAP, PAR_CTRLSCOREMAP, PAR_CTRLIMPOBJ,
        PAR_NIMPOBJ, PAR_CTRLOBJSCORES, PAR_NCASES, PAR_NCASETYPES, PAR_NMODELS, PAR_CTRLVEL, PAR_NRUNS, **PAR_CASERULE, PAR_SCORE, PAR_PROB,
        PAR_X, PAR_CTRLCDF, *PAR_MID, *PAR_MTYPE, *CASE_NAME, *CASE_TYPE, *CASE_S, *IMP_NAME, *IMP_TYPE, *IMP_S, *CDFO_O, *PDFO_O, *CDFL_O,
        *PDFL_O, SUBSET_MIN, SUBSET_MAX, *ris_obj, i, j, k, l, t, u, x, y, caseid, ncase, ncasetype, casetype, model, model_min, model_max,
        nimpobj, imparea, xint, releasemap, releasemap2, casemax, ncases, cdfid, pdfid, ctrl_aoi, nruns;

    double PAR_NRUNS0, PAR_SMIN, *PAR_RUNUPMAX, *PAR_FSLOPE, *PAR_FDIR, *PAR_SEG, *PAR_SCTRL,
        *PAR_MPAR1, *PAR_MPAR2, *PAR_MPAR3, *CASE_V, *CASE_Q, *CASE_PROB, *CASE_XR, *CASE_YR, *CASE_XS, *CASE_YS, *CDFO, *PDFO, *CDFL, *PDFL,
        deltas, prob, *aschdr, pdfomax, pdflmax;

    double CONST_PI = 3.14159;

    FILE *fparam, *fcases, *fobjscores, *fcdfo, *fcdfl, *fpdfo, *fpdfl, *fsubset;
    time_t timer;

    #ifdef WITHGRASS

        SEGMENT seg_aoi, seg_elev, seg_flowdir, seg_releasemap, seg_releasemap2, seg_impobj, seg_imparea, seg_scoremap, seg_probmap;

        struct GModule *module;
        struct Cell_head cellhd;

    #else

        int **seg_aoi, **seg_flowdir, **seg_releasemap, **seg_releasemap2, **seg_impobj, **seg_imparea, **seg_scoremap;
        double **seg_elev, **seg_probmap;

    #endif

    struct ico sico;

    #ifdef WITHGRASS


    // INITIALIZING GIS FUNCTIONS -------------------------------------------------------------


        G_gisinit( argv[0] );
        module = G_define_module();
        G_get_set_window( &cellhd );


// PREPARING ENVIRONMENT ----------------------------------------------------------------------


        sprintf(wkdir, "%s/%s/.tmp/rtemp", G_location_path(),G_mapset()); // defining working directory
        G_init_tempfile(); // initializing generation of temporary files
        sico.NSEGS = 1;
        sico.NSEGCR = 16;
        xint=10;
        yint=getenv("XINT");
        xint=atoi(yint);

    #else

        if ( argc > 1 ) sprintf( wkdir, "%s", argv[1] ); // defining working directory
        else {
            printf( "ERROR: Please provide a parameter file.\n" );
            exit( EXIT_SUCCESS );
        }
        xint=1;

    #endif


// READING PARAMETER FILE ---------------------------------------------------------------------


    sprintf(path, "%s/param%d.txt", wkdir, xint); // writing name of parameter file to string
    fparam=fopen(path, "r"); // opening parameter file

    if( !fparam ) { // testing whether parameter file exists:
        printf( "ERROR: Unable to open parameter file: '%s'\n", path );
        exit( EXIT_SUCCESS );
    }

    sico.MAINMAPSET = fcparam ( fparam ); // name of main mapset

    PAR_AFLAG = fiparam ( fparam ); // control for average of impact probability
    PAR_BACK = fiparam ( fparam ); // control for back-calculation of observation
    PAR_CONN = fiparam ( fparam ); // control for connectivity
    PAR_MULT = fiparam ( fparam ); // control for multiple model runs
    PAR_PROB = fiparam ( fparam ); // control for probabilities
    PAR_SCORE = fiparam ( fparam ); // control for scores
    PAR_SWALK = fiparam ( fparam ); // control for separate random walks for each model
    PAR_X = fiparam ( fparam ); // control for release from pixels

    PAR_PREFIX = fcparam ( fparam ); // prefix for output
    PAR_FILEPATH = fcparam ( fparam ); // path for output files
    PAR_MAPPATH = fcparam ( fparam ); // path for output files

    PAR_AOI = fcparam ( fparam ); // area of interest raster map


// SETTING EXTENT AND CELL SIZE ---------------------------------------------------------------


    #ifdef WITHGRASS

        sico.CSZ=cellhd.ew_res;
        sico.BDWEST=cellhd.west;
        sico.BDNORTH=cellhd.north;
        sico.N=cellhd.cols;
        sico.M=cellhd.rows;
        sico.UDM = -9999.0;
        sico.UDP = 9999.0;

        seg_aoi = finrasti( PAR_AOI, sico );

    #else

        aschdr = finaschdr( PAR_AOI, sico );

        sico.N = (int)aschdr[0]; // number of pixels in y direction (W-E)
        sico.M = (int)aschdr[1]; // number of pixels in x direction (S-N)
        sico.CSZ = aschdr[4]; // pixel size
        sico.BDWEST = aschdr[2]; // western boundary
        sico.BDSOUTH = aschdr[3]; // southern boundary
        sico.BDNORTH = aschdr[3] + sico.M * sico.CSZ; // northern boundary
        sico.UDM = aschdr[5]; // no data value
        sico.UDP = aschdr[5] * -1;

        free(aschdr);
        seg_aoi = finascvali( PAR_AOI, sico );

    #endif


// CONTINUING READING PARAMETERS --------------------------------------------------------------


    PAR_ELEV = fcparam ( fparam ); // elevation raster map

    #ifdef WITHGRASS

        seg_elev = finrastd( PAR_ELEV, sico );

    #else

        seg_elev = finascvald( PAR_ELEV, sico );

    #endif

    PAR_CTRLFLOWDIR = fiparam ( fparam ); // control for flow direction raster map
    if ( PAR_CTRLFLOWDIR == 1 ) {
        PAR_FLOWDIR = fcparam ( fparam ); // flow direction raster map

        #ifdef WITHGRASS

            seg_flowdir = finrasti( PAR_FLOWDIR, sico );

        #else

            seg_flowdir = finascvali( PAR_FLOWDIR, sico );

        #endif
    }

    PAR_CTRLPROBMAP = fiparam ( fparam ); // control for release probability raster map
    if ( PAR_CTRLPROBMAP == 1 ) {
        PAR_PROBMAP = fcparam ( fparam ); // release probability raster map

        #ifdef WITHGRASS

            seg_probmap = finrastd( PAR_PROBMAP, sico );

        #else

            seg_probmap = finascvald( PAR_PROBMAP, sico );

        #endif
    }

    PAR_CTRLRELEASEMAP = fiparam ( fparam ); // control for release area raster map
    if ( PAR_CTRLRELEASEMAP == 1 ) {
        PAR_RELEASEMAP = fcparam ( fparam ); // release area raster map

        #ifdef WITHGRASS

            seg_releasemap = finrasti( PAR_RELEASEMAP, sico );

        #else

            seg_releasemap = finascvali( PAR_RELEASEMAP, sico );

        #endif
    }

    #ifdef WITHGRASS

        if ( PAR_X == 1 ) Segment_open (&seg_releasemap2, G_tempfile(), sico.M, sico.N, sico.NSEGCR, sico.NSEGCR, sizeof(int), sico.NSEGS);

    #else

        if ( PAR_X == 1 ) seg_releasemap2 = alloc_imatrix( sico.M, sico.N );

    #endif

    PAR_CTRLSCOREMAP = fiparam ( fparam ); // control for release hazard indicator score raster map
    if ( PAR_CTRLSCOREMAP == 1 ) {
        PAR_SCOREMAP = fcparam ( fparam ); // release hazard indicator score raster map

        #ifdef WITHGRASS

            seg_scoremap = finrasti( PAR_SCOREMAP, sico );

        #else

            seg_scoremap = finascvali( PAR_SCOREMAP, sico );

        #endif
    }

    PAR_CTRLIMPOBJ = fiparam ( fparam ); // control for impact objects raster map
    if ( PAR_CTRLIMPOBJ == 1 ) {
        PAR_IMPOBJ = fcparam ( fparam ); // impact objects raster map

        #ifdef WITHGRASS

            seg_impobj = finrasti( PAR_IMPOBJ, sico );

        #else

            seg_impobj = finascvali( PAR_IMPOBJ, sico );

        #endif
    }

    PAR_CTRLOBJSCORES = fiparam ( fparam ); // control for object scores file
    if ( PAR_CTRLOBJSCORES == 1 ) PAR_NIMPOBJ = fiparam ( fparam ); // number of impact objects

    if ( PAR_BACK == 1 ) {
        PAR_IMPAREA = fcparam ( fparam ); // impact area raster map

        #ifdef WITHGRASS

            seg_imparea = finrasti( PAR_IMPAREA, sico );
        #else

            seg_imparea = finascvali( PAR_IMPAREA, sico );

        #endif
    }

    PAR_NCASES = fiparam ( fparam ); // number of cases
    PAR_NMODELS = fiparam ( fparam ); // number of models
    PAR_CTRLVEL = fiparam ( fparam ); // control for velocity

    PAR_NRUNS0 = fdparam ( fparam ); // number of random walks for each case

    if ( PAR_NRUNS0 == -9999 ) PAR_NRUNS = -9999;
    else PAR_NRUNS = ( int )( pow( 10, PAR_NRUNS0 ) + 0.5 );
    nruns = PAR_NRUNS;

    PAR_SMIN = fdparam ( fparam ); // minimum travel distance

    PAR_RUNUPMAX = (double*) calloc( PAR_NMODELS, sizeof(double));
    PAR_FSLOPE = (double*) calloc( PAR_NMODELS, sizeof(double));
    PAR_FDIR = (double*) calloc( PAR_NMODELS, sizeof(double));
    PAR_SEG = (double*) calloc( PAR_NMODELS, sizeof(double));
    PAR_SCTRL = (double*) calloc( PAR_NMODELS, sizeof(double));


    if ( PAR_SWALK == 0 ) {

        PAR_SCTRL[0] = fdparam ( fparam ); // control distance for maintenance of direction
        PAR_SEG[0] = fdparam ( fparam ); // segment length
        PAR_RUNUPMAX[0] = fdparam ( fparam ); // maximum runup
        PAR_FSLOPE[0] = fdparam ( fparam ); // factor for weighting slope
        PAR_FDIR[0] = fdparam ( fparam ); // factor for weighing maintenance of direction
    }

    PAR_MID = (int*) calloc( PAR_NMODELS, sizeof(int));
    PAR_MTYPE = (int*) calloc( PAR_NMODELS, sizeof(int));
    PAR_MPAR1 = (double*) calloc( PAR_NMODELS, sizeof(double));
    PAR_MPAR2 = (double*) calloc( PAR_NMODELS, sizeof(double));
    PAR_MPAR3 = (double*) calloc( PAR_NMODELS, sizeof(double));

    for ( model = 0; model < PAR_NMODELS; model++ ) { // loop over all models:

        PAR_MID[model] = fiparam ( fparam ); // model id
        PAR_MTYPE[model] = fiparam ( fparam ); // type of model
        PAR_MPAR1[model] = fdparam ( fparam ); // first model parameter
        PAR_MPAR2[model] = fdparam ( fparam ); // second model parameter
        PAR_MPAR3[model] = fdparam ( fparam ); // third model parameter

        if ( PAR_SWALK == 1 ) {

            PAR_SCTRL[model] = fdparam ( fparam ); // control distance for maintenance of direction
            PAR_SEG[model] = fdparam ( fparam ); // segment length
            PAR_RUNUPMAX[model] = fdparam ( fparam ); // maximum runup
            PAR_FSLOPE[model] = fdparam ( fparam ); // factor for weighting slope
            PAR_FDIR[model] = fdparam ( fparam ); // factor for weighing maintenance of direction
        }
    }

    PAR_NCASETYPES = fiparam ( fparam ); // number of case types
    if ( PAR_NCASETYPES > 0 ) {

        PAR_CASERULE = alloc_imatrix( PAR_NCASETYPES+1, PAR_NMODELS );
        for ( ncasetype = 0; ncasetype < PAR_NCASETYPES; ncasetype++ ) { // loop over all case types:

            casetype = fiparam ( fparam ); // case type
            for ( model = 0; model < PAR_NMODELS; model++ ) PAR_CASERULE[casetype][model] = fdparam ( fparam ); // case rule for model
        }
    }

    PAR_CTRLCDF = fiparam ( fparam ); // control for cdf file
    fclose(fparam); // closing parameter file


// READING CDF, SUBSETTING, CASE AND OBJECT SCORES FILES --------------------------------------


    if ( PAR_CTRLCDF == 1 || PAR_CTRLCDF == 3 ) { // file with cumulative density function of angle of reach:

        CDFO_O = (int*) calloc( 2001, sizeof(int));
        CDFO = (double*) calloc( 2001, sizeof(double));

        if ( PAR_MULT == 0 ) sprintf(path, "%s/cdfo.txt", wkdir); // writing name of cdf file to string
        else sprintf(path, "%s/cdfo%i.txt", wkdir, xint);
        fcdfo=fopen(path, "r"); // opening cdf file

        for ( cdfid = 0; cdfid < 2001; cdfid++ ) { // loop over all cdf values:

            CDFO_O[cdfid] = (int)(1000 * fdparam ( fcdfo )); // angle of reach
            CDFO[cdfid] = fdparam ( fcdfo ); // cdf
        }
        fclose(fcdfo); // closing cdf file

        pdfomax = 0;
        PDFO_O = (int*) calloc( 2001, sizeof(int));
        PDFO = (double*) calloc( 2001, sizeof(double));

        if ( PAR_MULT == 0 ) sprintf(path, "%s/pdfo.txt", wkdir); // writing name of pdf file to string
        else sprintf(path, "%s/pdfo%i.txt", wkdir, xint);
        fpdfo=fopen(path, "r"); // opening pdf file

        for ( pdfid = 0; pdfid < 2001; pdfid++ ) { // loop over all pdf values:

            PDFO_O[pdfid] = (int)(1000 * fdparam ( fpdfo )); // angle of reach
            PDFO[pdfid] = fdparam ( fpdfo ); // pdf
            if ( PDFO[pdfid] > pdfomax ) pdfomax = PDFO[pdfid]; // updating maximum of pdf
        }
        fclose(fpdfo); // closing pdf file
    }

    if ( PAR_CTRLCDF == 2 || PAR_CTRLCDF == 3 ) { // file with cumulative density function of maximum travel distance:

        CDFL_O = (int*) calloc( 2001, sizeof(int));
        CDFL = (double*) calloc( 2001, sizeof(double));

        if ( PAR_MULT == 0 ) sprintf(path, "%s/cdfl.txt", wkdir); // writing name of cdf file to string
        else sprintf(path, "%s/cdfl%i.txt", wkdir, xint);
        fcdfl=fopen(path, "r"); // opening cdf file

        for ( cdfid = 0; cdfid < 2001; cdfid++ ) { // loop over all cdf values:

            CDFL_O[cdfid] = (int)(1000 * fdparam ( fcdfl )); // maximum travel distance
            CDFL[cdfid] = fdparam ( fcdfl ); // cdf
        }
        fclose(fcdfl); // closing cdf file

        pdflmax = 0;
        PDFL_O = (int*) calloc( 2001, sizeof(int));
        PDFL = (double*) calloc( 2001, sizeof(double));

        if ( PAR_MULT == 0 ) sprintf(path, "%s/pdfl.txt", wkdir); // writing name of pdf file to string
        else sprintf(path, "%s/pdfl%i.txt", wkdir, xint);
        fpdfl=fopen(path, "r"); // opening pdf file

        for ( pdfid = 0; pdfid < 2001; pdfid++ ) { // loop over all pdf values:

            PDFL_O[pdfid] = (int)(1000 * fdparam ( fpdfl )); // maximum travel distance
            PDFL[pdfid] = fdparam ( fpdfl ); // pdf
            if ( PDFL[pdfid] > pdflmax ) pdflmax = PDFL[pdfid]; // updating maximum of pdf
        }
        fclose(fpdfl); // closing pdf file
    }

    if ( PAR_MULT == 1 && PAR_PROB == 1 && PAR_NCASES > 0 ) { // file for subsetting release file:

        sprintf(path, "%s/cdfthr%i.txt", wkdir, xint);
        fsubset=fopen(path, "r"); // opening file with subset parameters

        SUBSET_MIN = fiparam ( fsubset ); // first case to apply
        SUBSET_MAX = fiparam ( fsubset ); // last case to apply

        fclose(fsubset); // closing subset file
    }
    else { // if file is not provided, applying entire set of cases:
        SUBSET_MIN = 0; // first case to apply
        SUBSET_MAX = 999999; // last case to apply
    }

    if ( PAR_NCASES > 0 ) { // if release file is provided, reading release file:

        ncases = PAR_NCASES + 1; // applying number of cases from parameter input
        sprintf(path, "%s/releasefile.txt", wkdir); // writing name of cases file to string
        fcases=fopen(path, "r"); // opening cases file

        CASE_NAME = (int*) calloc( ncases, sizeof(int));
        CASE_TYPE = (int*) calloc( ncases, sizeof(int));
        CASE_V = (double*) calloc( ncases, sizeof(double));
        CASE_Q = (double*) calloc( ncases, sizeof(double));
        CASE_S = (int*) calloc( ncases, sizeof(int));
        CASE_PROB = (double*) calloc( ncases, sizeof(double));
        CASE_XR = (double*) calloc( ncases, sizeof(double));
        CASE_YR = (double*) calloc( ncases, sizeof(double));
        CASE_XS = (double*) calloc( ncases, sizeof(double));
        CASE_YS = (double*) calloc( ncases, sizeof(double));

        for ( caseid = 1; caseid < ncases; caseid++ ) { // loop over all cases:

            CASE_NAME[caseid] = fiparam ( fcases ); // case name
            CASE_TYPE[caseid] = fiparam ( fcases ); // case type
            CASE_V[caseid] = fdparam ( fcases ); // case magnitude
            CASE_Q[caseid] = fdparam ( fcases ); // case peak discharge
            CASE_S[caseid] = fiparam ( fcases ); // case release hazard indicator score
            CASE_PROB[caseid] = fdparam ( fcases ); // case release probability
            CASE_XR[caseid] = fdparam ( fcases ); // x coordinate of release point
            CASE_YR[caseid] = fdparam ( fcases ); // y coordinate of release point
            CASE_XS[caseid] = fdparam ( fcases ); // x coordinate of starting point
            CASE_YS[caseid] = fdparam ( fcases ); // y coordinate of starting point
        }
        fclose(fcases); // closing release file
    }
    if ( PAR_X == 1 ) { // if no case file is provided:

        ncases = 0; // initializing number of cases
        caseid = 0; // initializing case id
        for ( x=0; x<sico.M; x++ ) { // loop over all pixels:
            for ( y=0; y<sico.N; y++ ) {

                if ( PAR_CTRLPROBMAP == 0 && PAR_CTRLSCOREMAP == 0 && PAR_BACK == 0 && ( PAR_CTRLRELEASEMAP == 1 || PAR_NCASES > 0 ))

                    #ifdef WITHGRASS

                        Segment_get(&seg_releasemap, &releasemap2, x, y ); // release map

                    #else

                        releasemap2 = seg_releasemap[x][y]; // release map

                    #endif

                else { // if release probability or score is defined, assigning separate case ids to each pixel within area of interest:

                    #ifdef WITHGRASS

                        Segment_get(&seg_aoi, &ctrl_aoi, x, y ); // area of interest
                        if ( PAR_BACK == 1 ) Segment_get(&seg_releasemap, &releasemap2, x, y ); // release map

                    #else

                        ctrl_aoi = seg_aoi[x][y]; // area of interest
                        if ( PAR_BACK == 1 ) releasemap2 = seg_releasemap[x][y]; // release map


                    #endif

                    if ( PAR_CTRLSCOREMAP == 1 ) Segment_get(&seg_scoremap, &releasemap2, x, y ); // score map

                    if (( PAR_BACK == 0 && PAR_CTRLSCOREMAP == 0 && ctrl_aoi > 0 ) 
                        || ( PAR_BACK == 1 && releasemap2 > 0 ) || ( PAR_CTRLSCOREMAP == 1 && releasemap2 > 0 )) {

                        caseid += 1; // increasing case id
                        releasemap2 = caseid; 
                    }
                    else releasemap2 = sico.UDM;
                }

                #ifdef WITHGRASS

                    Segment_put(&seg_releasemap2, &releasemap2, x, y ); // applying case id to pixel of release map

                #else

                    seg_releasemap2[x][y] = releasemap2; // applying case id to pixel of release map

                #endif

                if ( releasemap2 > ncases ) ncases = releasemap2; // updating number of cases
            }
        }

        #ifdef WITHGRASS

            Segment_flush(&seg_releasemap2);

        #endif

        ncases += 1; // finalizing number of cases
    }

    if ( PAR_CTRLOBJSCORES == 1 ) { // object scores file:

        IMP_NAME = (int*) calloc( PAR_NIMPOBJ, sizeof(int));
        IMP_TYPE = (int*) calloc( PAR_NIMPOBJ, sizeof(int));
        IMP_S = (int*) calloc( PAR_NIMPOBJ, sizeof(int));
        ris_obj = (int*) calloc( PAR_NIMPOBJ, sizeof(int));

        sprintf(path, "%s/objscores.txt", wkdir); // writing name of object scores file to string
        fobjscores=fopen(path, "r"); // opening object scores file

        for ( nimpobj = 0; nimpobj < PAR_NIMPOBJ; nimpobj++ ) {

            IMP_NAME[nimpobj] = fiparam ( fobjscores ); // name of impact object
            IMP_TYPE[nimpobj] = fiparam ( fobjscores ); // type of impact object
            IMP_S[nimpobj] = fiparam ( fobjscores ); // object score
            ris_obj[nimpobj] = 0;
        }
        fclose(fobjscores); // closing object scores file
    }


// DECLARING WORKING AND OUTPUT VARIABLES -----------------------------------------------------


    char mout[1000], miis[1000], msmov[1000], mzmov[1000], momega[1000], mxv[1000], mhis[1000], mris_impobj[1000],
        mimpprobr[1000], mrelprobr[1000], mimpquant[1000], msmovi[1000], mzmovi[1000], momegai[1000], mconn[1000];

    int kimpact, ctr_run, ctr_run_test, ctr_run_seg, xstart, ystart, xf, yf, xseg, yseg, ctr_failalloc, fdir, randmax, randango, randangl,
        *impquant, impquantold, impquantnew, ctrl_M[PAR_NMODELS], fdir_prev, ctrl_valid, ctrl_valid2, ctrl_jimpact, iimpact, jimpact,
        ctrl_releasemap, ctrl_impobj, ctrl_impbar, ctrlall, his_test, *his, ris, iisold, iMold, iMnew, hisold, xstop, ystop, xrun, yrun,
        ctrl_run, kall, ctr_kall, ctr_models, ctr_rscore, rid, xprob, yprob, xstart2, ystart2, aoi_f, xprob_test, yprob_test,
        casename, casename3, cases, nimpact_all, fdirtest, ctrl_impact, ctrl_mpar1, lconn0, lconn;

    double pcomp, elev, elev_release, elev_start, elev_f, elev_segstart, elev_segstart0, randnum, elev_adj, elev_ctrl_prev, elev_seg_prev,
        z_mov, runupmax, direction_factor, dist_ctrl_app, dist_ctrl_prevapp, z_movmax, smov_test, smov_outold, zmov_outold, omega_outold,
        *s_mov_out, *z_mov_out, *omega_out, *xxv, omega_det, s_mov_det, s_mov_segstart0, xdalpha, xv, xv1, xv2, xzeta, xeta,
        impprob_test, *impprob, *impprobr, impprobr_test, impprobr_testo, impprobr_testl, impprobrold, impprobrnew,  impprobold, s_mov_seg,
        s_mov_segstart, s_mov_segstart00, relprobrnew, relprobrold, casev, caseq, caseprob, betax, betay, alphaprev, alphadiff1, alphadiff2,
        alphadiff, smovi, zmovi, omegai, xf_segstart00, yf_segstart00, randprob, z_movprev, omegatop, omegaup, omegatrans;

    FILE *f_summary, *f_backfile, *f_impobj, *f_nout, *f_conn;

    #ifdef WITHGRASS

        SEGMENT seg_iM[PAR_NMODELS], seg_conn, seg_iis, seg_smov, seg_zmov, seg_omega, seg_xv, seg_his, seg_ris_impobj, seg_impprob, seg_relprobr,
            seg_impprobr, seg_impquant, seg_smovi, seg_zmovi, seg_omegai;

    #else

        int ***seg_iM, **seg_conn, **seg_iis, **seg_his, **seg_ris_impobj, **seg_impquant;
        double **seg_smov, **seg_zmov, **seg_omega, **seg_xv, **seg_impprob, **seg_relprobr, **seg_impprobr,
            **seg_smovi, **seg_zmovi, **seg_omegai;

    #endif

    int *xadj = (int*) calloc( 9, sizeof(int));
    int *yadj = (int*) calloc( 9, sizeof(int));
    //int *fdirtestd = (int*) calloc( 9, sizeof(int));
    double *alphatestd = (double*) calloc( 9, sizeof(double));

    int *whtslope = (int*) calloc( 9, sizeof(int));
    double *xalpha = (double*) calloc( 2, sizeof(double));

    double *s_mov = (double*) calloc( 10000, sizeof(double));
    double *omega = (double*) calloc( 10000, sizeof(double));

    double *s_movmax = (double*) calloc( PAR_NMODELS, sizeof(double));
    double *omega_stop = (double*) calloc( PAR_NMODELS, sizeof(double));
    double *omega_test = (double*) calloc( PAR_NMODELS, sizeof(double));

    int *xf_run = (int*) calloc( 10000, sizeof(int));
    int *yf_run = (int*) calloc( 10000, sizeof(int));

    int *ximpact = (int*) calloc( 100000, sizeof(int));
    int *yimpact = (int*) calloc( 100000, sizeof(int));
    int *iis = (int*) calloc( 100000, sizeof(int));
    int **iM = alloc_imatrix( PAR_NMODELS, 100000 );

    s_mov_out = (double*) calloc( 100000, sizeof(double));
    z_mov_out = (double*) calloc( 100000, sizeof(double));
    omega_out = (double*) calloc( 100000, sizeof(double));
    if ( PAR_CTRLVEL == 1 ) xxv = (double*) calloc( 100000, sizeof(double));
    impquant = (int*) calloc( 100000, sizeof(int));

    if ( PAR_PROB == 1 ) {
        impprobr = (double*) calloc( 100000, sizeof(double));
        impprob = (double*) calloc( 100000, sizeof(double));
    }
    if ( PAR_SCORE == 1 ) his = (int*) calloc( 100000, sizeof(int));


// PREPARING OUTPUT FILES ---------------------------------------------------------------------


    if ( PAR_MULT == 0 ) {
        sprintf(paths, "%s%s_summary.txt", PAR_FILEPATH, PAR_PREFIX); // summary file
        if ( PAR_CTRLOBJSCORES == 1 ) sprintf(patho, "%s%s_ris.txt", PAR_FILEPATH, PAR_PREFIX); // risk indicator score file
    }
    else sprintf(paths, "%s%s_summary%d.txt", PAR_FILEPATH, PAR_PREFIX, xint); // for multiple model runs, only summary file is relevant
    if ( PAR_BACK == 1 ) sprintf(pathd, "%s%s_backfile.txt", PAR_FILEPATH, PAR_PREFIX); // back-calculation file

    f_summary=fopen(paths, "w"); // opening files:
    if ( PAR_BACK == 1 ) f_backfile=fopen(pathd, "w");
    if ( PAR_CTRLIMPOBJ == 1 && PAR_CTRLOBJSCORES == 1 ) f_impobj=fopen(patho, "w");

    if ( PAR_MULT == 1 ) {
        sprintf(path, "%s%s_parout%d.txt", PAR_FILEPATH, PAR_PREFIX, xint); // file with information on model run:
        f_nout=fopen(path, "w");
    }
    if ( PAR_CONN == 1 && PAR_NCASES > 0 ) {
        sprintf(pathc, "%s%s_conn.txt", PAR_FILEPATH, PAR_PREFIX); // file with connectivity results
        f_conn=fopen(pathc, "w");
        fprintf(f_conn, "CASE\tXR\tYR\tSCONN\n", casename, x, y, lconn);
    }


// PREPARING SEGMENT FILES FOR OUTPUT RASTER MAPS ---------------------------------------------


    #ifdef WITHGRASS

        Segment_open (&seg_iis, G_tempfile(), sico.M, sico.N, sico.NSEGCR, sico.NSEGCR, sizeof(int), sico.NSEGS);
        Segment_open (&seg_impquant, G_tempfile(), sico.M, sico.N, sico.NSEGCR, sico.NSEGCR, sizeof(int), sico.NSEGS);

        for ( model = 0; model < PAR_NMODELS; model++ )
            Segment_open (&seg_iM[model], G_tempfile(), sico.M, sico.N, sico.NSEGCR, sico.NSEGCR, sizeof(int), sico.NSEGS);

        if ( PAR_CONN == 1 ) Segment_open (&seg_conn, G_tempfile(), sico.M, sico.N, sico.NSEGCR, sico.NSEGCR, sizeof(int), sico.NSEGS);

        if ( PAR_MULT == 0 ) {

            Segment_open (&seg_smov, G_tempfile(), sico.M, sico.N, sico.NSEGCR, sico.NSEGCR, sizeof(double), sico.NSEGS);
            Segment_open (&seg_zmov, G_tempfile(), sico.M, sico.N, sico.NSEGCR, sico.NSEGCR, sizeof(double), sico.NSEGS);
            Segment_open (&seg_omega, G_tempfile(), sico.M, sico.N, sico.NSEGCR, sico.NSEGCR, sizeof(double), sico.NSEGS);

            if ( PAR_CTRLVEL == 1 )
                Segment_open (&seg_xv, G_tempfile(), sico.M, sico.N, sico.NSEGCR, sico.NSEGCR, sizeof(double), sico.NSEGS);

            if ( PAR_CTRLIMPOBJ == 1 ) {
                Segment_open (&seg_ris_impobj, G_tempfile(), sico.M, sico.N, sico.NSEGCR, sico.NSEGCR, sizeof(int), sico.NSEGS);
                Segment_open (&seg_smovi, G_tempfile(), sico.M, sico.N, sico.NSEGCR, sico.NSEGCR, sizeof(double), sico.NSEGS);
                Segment_open (&seg_zmovi, G_tempfile(), sico.M, sico.N, sico.NSEGCR, sico.NSEGCR, sizeof(double), sico.NSEGS);
                Segment_open (&seg_omegai, G_tempfile(), sico.M, sico.N, sico.NSEGCR, sico.NSEGCR, sizeof(double), sico.NSEGS);
            }
        }

        if ( PAR_PROB == 1 ) {

            if ( PAR_MULT == 0 )
                Segment_open (&seg_relprobr, G_tempfile(), sico.M, sico.N, sico.NSEGCR, sico.NSEGCR, sizeof(double), sico.NSEGS);
            Segment_open (&seg_impprobr, G_tempfile(), sico.M, sico.N, sico.NSEGCR, sico.NSEGCR, sizeof(double), sico.NSEGS);
            if ( PAR_MULT == 0 )
                Segment_open (&seg_impprob, G_tempfile(), sico.M, sico.N, sico.NSEGCR, sico.NSEGCR, sizeof(double), sico.NSEGS);
        }

        if ( PAR_SCORE == 1 ) Segment_open (&seg_his, G_tempfile(), sico.M, sico.N, sico.NSEGCR, sico.NSEGCR, sizeof(int), sico.NSEGS);

    #else

        seg_iis = alloc_imatrix( sico.M, sico.N );
        seg_impquant = alloc_imatrix( sico.M, sico.N );
        seg_iM = alloc_imatrix3( PAR_NMODELS, sico.M, sico.N );

        if ( PAR_CONN == 1 ) seg_conn = alloc_imatrix( sico.M, sico.N );

        if ( PAR_MULT == 0 ) {

            seg_smov = alloc_dmatrix( sico.M, sico.N );
            seg_zmov = alloc_dmatrix( sico.M, sico.N );
            seg_omega = alloc_dmatrix( sico.M, sico.N );
            if ( PAR_CTRLVEL == 1 ) seg_xv = alloc_dmatrix( sico.M, sico.N );
            if ( PAR_CTRLIMPOBJ == 1 ) {
                seg_ris_impobj = alloc_imatrix( sico.M, sico.N );
                seg_smovi = alloc_dmatrix( sico.M, sico.N );
                seg_zmovi = alloc_dmatrix( sico.M, sico.N );
                seg_omegai = alloc_dmatrix( sico.M, sico.N );
            }
        }

        if ( PAR_PROB == 1 ) {

            if ( PAR_MULT == 0 ) seg_relprobr = alloc_dmatrix( sico.M, sico.N );
            seg_impprobr = alloc_dmatrix( sico.M, sico.N );
            if ( PAR_MULT == 0 ) seg_impprob = alloc_dmatrix( sico.M, sico.N );
        }

        if ( PAR_SCORE == 1 ) seg_his = alloc_imatrix( sico.M, sico.N );

    #endif


// PRINTING HEADERS OF OUTPUT FILES AND DISPLAY -----------------------------------------------


    fprintf(f_summary, "ID");
    printf("\n   ID\tSTATUS");

    for ( model = 0; model < PAR_NMODELS; model++ ) {
        fprintf(f_summary, "\tLMAX%i\tOMEGAT%i", PAR_MID[model], PAR_MID[model]);
        printf("\tLMAX%i\tOMEGAT%i", PAR_MID[model], PAR_MID[model]);
    }
    fprintf(f_summary, "\tRMAX\tAREA\n");
    printf("\tRMAX\tAREA\n");

    if ( PAR_BACK == 1 ) fprintf(f_backfile, "ID\tCASE\tLMAX\tOMEGAT\n");


// PREPARING ROUTING --------------------------------------------------------------------------


    timer=time(NULL); // timer for randomization
    srand(timer); // seed for random number

    rid = 0; // initializing counter for reconstruction cases
    xprob = 0; yprob = -1; // initializing start coordinates if all valid pixels are considered for release

    for ( caseid = 1; caseid < ncases; caseid++ ) { // loop over all cases:

        impquant[0] = 0; // initializing number of cases possibly affecting the pixel
        if ( PAR_PROB == 1 ) impprobr[0] = 0;  // inizializing impact probability
        jimpact = 0; // resetting counter for number of affected pixels
        ctr_kall = 0; // initializing counter for the total number of random walks for the case
        ctr_failalloc = 0; // resetting counter for unsuccessful routing attempts
        runupmax = 0; // resetting maximum runup height

        for ( model = 0; model < PAR_NMODELS; model++ ) {
            s_movmax[model] = sico.UDM; omega_stop[model] = sico.UDP; // resetting maximum travel distance and angle of reach
        }

        if ( PAR_NCASES > 0 ) {
            casename = CASE_NAME[caseid]; // case name
            casetype = CASE_TYPE[caseid]; // case type
            casev = CASE_V[caseid]; // case volume
            caseq = CASE_Q[caseid]; // case peak discharge
            cases = CASE_S[caseid]; // case release score
            caseprob = CASE_PROB[caseid]; // case release probability
        }
        else {
            casename = caseid; // case name
            casetype = 1; // case type
            casev = sico.UDM; // case volume
            caseq = sico.UDM; // case peak discharge
            cases = sico.UDM; // case score
            caseprob = sico.UDM; // case release probability
        }

        if ( PAR_X == 1 && PAR_CTRLRELEASEMAP == 1 && PAR_CTRLPROBMAP == 0 && PAR_CTRLSCOREMAP == 0 && PAR_BACK == 0 ) {
            // for start from pixels defined by the release map:

            kall = 0; // initializing total number of random walks for the case
            xstart2 = 0; ystart2 = 0; xstop = sico.M - 1; ystop = sico.N - 1; // initializing release variables

            for ( xrun=xstart2; xrun<xstop; xrun++ ) { // loop over all pixels:
                for ( yrun=ystart2; yrun<ystop; yrun++ ) {

                    #ifdef WITHGRASS

                        Segment_get(&seg_releasemap2, &releasemap2, xrun, yrun); // release map (case id)

                    #else

                        releasemap2 = seg_releasemap2[xrun][yrun]; // release map (case id)

                    #endif

                    if ( releasemap2 == casename ) kall += PAR_NRUNS; // updating total number of random walks for the case
                }
            }
        }

        else if ( PAR_X == 1 ) { // if separate case ids were assigned to each pixel within area of interest:

            releasemap2 = sico.UDM; // initializing case id
            while ( releasemap2 != casename && xprob < sico.M-2 ) { // searching for the next valid pixel:

                if ( yprob > sico.N-2 ) { yprob = 0; xprob += 1; } else yprob += 1; // updating test variables

                #ifdef WITHGRASS

                    Segment_get(&seg_releasemap2, &releasemap2, xprob, yprob); // case id

                #else

                    releasemap2 = seg_releasemap2[xprob][yprob]; // case id

                #endif

            }
            xstart2 = xprob; ystart2 = yprob; xstop = xprob + 1; ystop = yprob + 1; // initializing release variables
            kall = PAR_NRUNS; // total number of random walks for the case
        }

        else { xstart2 = 0; ystart2 = 0; xstop = 1; ystop = 1; kall = PAR_NRUNS; }
            // release variables and total number of random walks for the case if a release file is provided

        for ( xrun=xstart2; xrun<xstop; xrun++ ) { for ( yrun=ystart2; yrun<ystop; yrun++ ) { // loop over all release pixels of the case:

            #ifdef WITHGRASS

                if ( PAR_CONN == 1 ) Segment_get(&seg_conn, &lconn, x, y); // connectivity

            #else

                if ( PAR_CONN == 1 ) lconn = seg_conn[x][y]; // connectivity

            #endif

            if ( PAR_X == 1 ) { // for release from pixels:

                #ifdef WITHGRASS

                    Segment_get(&seg_releasemap2, &releasemap2, xrun, yrun); // release map (case id)

                #else

                    releasemap2 = seg_releasemap2[xrun][yrun]; // release map (case id)

                #endif

                if ( releasemap2 == casename && releasemap2 != sico.UDM ) { // if pixel is valid:

                    ctrl_run = 1; // setting control for starting random walks to positive
                    x = xrun; y = yrun; // coordinates of release point
                    xstart = xrun; ystart = yrun; // coordinates of start of random walk
                }
                else ctrl_run = 0; // if pixel is invalid, setting control for starting random walks to negative
            }

            else { // for release from coordinates defined by release file:

                ctrl_run = 1;

                x = (int)(( sico.BDNORTH - CASE_YR[caseid] ) / sico.CSZ + 0.5 ); // coordinates of release point
                y = (int)(( CASE_XR[caseid] - sico.BDWEST ) / sico.CSZ + 0.5 );

                xstart = (int)(( sico.BDNORTH - CASE_YS[caseid] ) / sico.CSZ + 0.5 ); // coordinates of start of random walk
                ystart = (int)(( CASE_XS[caseid] - sico.BDWEST ) / sico.CSZ + 0.5 );
            }

            if ( PAR_NCASES > 0 ) { // if release file is provided:
                if ( PAR_PROB == 1 ) prob = caseprob; else prob = 1; // release probability
                if ( PAR_SCORE == 1 ) ctr_rscore = cases; else ctr_rscore = 1; // release hazard indicator score

                if ( PAR_X == 0 ) { // if random walks are started from release file:

                    if ( SUBSET_MAX > SUBSET_MIN && ( casename >= SUBSET_MIN && casename <= SUBSET_MAX )) ctrl_run = 1;
                    else if ( SUBSET_MAX < SUBSET_MIN && ( casename >= SUBSET_MIN || casename <= SUBSET_MAX )) ctrl_run = 1;
                    else ctrl_run = 0; // verifying that case is in subset
                }
            }
            else { // if no release file is provided:

                #ifdef WITHGRASS

                    if ( PAR_CTRLPROBMAP == 1 ) { // release probability
                        Segment_get(&seg_probmap, &prob, xrun, yrun);
                        if (PAR_NRUNS == -9999 ) nruns = ( int )( prob * 100 ); // number of random walks
                    }
                    else prob = 1;

                    if ( PAR_SCORE == 1 ) Segment_get(&seg_scoremap, &ctr_rscore, xrun, yrun); else ctr_rscore = 1;
                        // release hazard indicator score

                #else

                    if ( PAR_CTRLPROBMAP == 1 ) {
                        prob = seg_probmap[xrun][yrun]; // release probability
                        if (PAR_NRUNS == -9999 ) nruns = ( int )( prob * 100 ); // number of random walks
                    }
                    else prob = 1; // release probability

                    if ( PAR_SCORE == 1 ) ctr_rscore = seg_scoremap[xrun][yrun]; // release hazard indicator score

                #endif

            }

            if ( PAR_BACK == 1 ) { // for back-calculation:
                if ( PAR_NCASES > 0 ) casename3 = casename;

                #ifdef WITHGRASS

                    else Segment_get(&seg_releasemap, &casename3, xrun, yrun); // release map (case id)

                #else

                    casename3 = seg_releasemap[xrun][yrun]; // release map (case id)

                #endif
            }

            #ifdef WITHGRASS

                if ( ctrl_run == 1 ) Segment_get(&seg_aoi, &ctrl_aoi, x, y); else ctrl_aoi = sico.UDM; // area of interest

            #else

                if ( ctrl_run == 1 ) ctrl_aoi = seg_aoi[x][y]; // area of interest

            #endif

            if ( ctrl_run == 1 && ( PAR_X == 0 || prob > 0 || ctr_rscore > 0 )
                && x >= 2 && x <= (sico.M-3) && y >= 2 && y <= (sico.N-3) && ctrl_aoi > 0 ) { // if release pixel supports routing:

                #ifdef WITHGRASS

                    Segment_get(&seg_elev, &elev_release, x, y); // elevation of release point

                    Segment_get(&seg_elev, &elev_start, xstart, ystart); // elevation of start of random walk

                #else

                    elev_release = seg_elev[x][y]; // elevation of release point
                    elev_start = seg_elev[xstart][ystart]; // elevation of start of random walk

                #endif

                if ( PAR_BACK == 1  ) { omega_det = sico.UDP; s_mov_det = sico.UDM; }
                    // resetting travel distance and angle of reach with respect to release pixel

                if ( PAR_CTRLIMPOBJ == 1 ) { smovi = sico.UDP * 100; zmovi = sico.UDP; omegai = sico.UDM; }
                    // resetting travel distance and angle of reach related to impact objects

                lconn = 0;


            // STARTING SERIES OF RANDOM WALKS ------------------------------------------------


                for ( k=0; k<nruns; k++ ) { // loop over defined number of random walks:

                    ctr_models = 0;
                    for ( model = 0; model < PAR_NMODELS; model++ ) { // activating models:

                        if ( PAR_NCASETYPES == 0 ) { // if no case rules are provided, activating all models
                            ctrl_M[model] = 1;
                            ctr_models += 1;
                        }
                        else if ( PAR_CASERULE[casetype][model] > 0 ) { // activating model according to case rule
                            ctrl_M[model] = 1;
                            ctr_models += 1;
                        }
                        else ctrl_M[model] = 0; // deactivating model according to case rule
                    }

                    ctr_kall+=1; // updating counter for number of random walks for case
                    pcomp = (double)100 * (double)ctr_kall / (double)kall; // per cent of case completed
                    if ( (int)pcomp % 1 == 0 ) printf ("   %i\t%.0f%%\r", casename, pcomp);

                    fdir_prev = sico.UDM; // resetting previous flow direction
                    //alphaprev = sico.UDM; // resetting previous flow direction
                    xf_run[0] = x; yf_run[0] = y; // setting initial coordinates to the release coordinates 
                    ximpact[0] = xstart; yimpact[0] = ystart; // storing start coordinates
                    s_mov[0] = pow ( pow(( xstart - x ) * sico.CSZ, 2 ) + pow(( ystart - y ) * sico.CSZ, 2 ), 0.5 );
                        // initializing travel distance
                    if ( PAR_PROB == 1 ) impprobr[0] = 1; // setting impact probability
                    impquant[0] = 1; // setting number of cases possibly affecting the pixel
                    z_mov = elev_release - elev_start; // initializing loss of elevation
                    z_movmax = 0; // resetting maximum drop of elevation
                    z_movprev = 0;
                    omegaup = 0;
                    omegatop = 0;
                    omegatrans = 0;
                    xf = xstart; yf = ystart; // initializing coordinates for direction
                    ctrl_aoi = 1; // resetting control for area of interest
                    ctrl_valid = 1; // resetting control for raster cell validity
                    ctrl_valid2 = 1; // resetting control for validity of all adjacent pixels
                    ctrl_impbar = 0; // resetting control for barrier
                    ctrlall = 1; // resetting number of active models
                    ctr_run = 0; // resetting number of step of random walk
                    xv1 = 0; // resetting flow velocity (only used with models of type 5)
                    xseg = xstart; yseg = ystart; // resetting segment start coordinates
                    s_mov_segstart = s_mov[0]; // resetting horizontal travel distance at start of segment
                    s_mov_segstart0 = 0; // resetting previous horizontal travel distance at start of segment
                    elev_segstart0 = elev_release; // resetting previous elevation at start of segment
                    xalpha[0] = 0; // initializing previous angle along path
                    ctrl_impact = 0; // initializing control for impact on impact object

                    ctrl_mpar1 = 1;
                    while ( ctrl_mpar1 == 1 && PAR_NRUNS0 == sico.UDM ) {

                        randmax=2000; // maximum for random number
                        randango=(rand()%randmax); // random number for angle of reach

                        randmax=2000; // maximum for random number
                        randangl=(rand()%randmax); // random number for maximum travel distance

                        if ( PAR_CTRLCDF == 1 ) {

                            randmax=(int)(1000*pdfomax); // maximum for random number
                            randprob=((double)(rand()%randmax))/1000; // random number for probability

                            if ( randprob < PDFO[randango] ) {
                                PAR_MPAR1[0] = (double)(PDFO_O[randango]) / 1000 * 180 / CONST_PI; // angle of reach
                                ctrl_mpar1 = 0;
                            }
                        }

                        else if ( PAR_CTRLCDF == 2 ) {

                            randmax=(int)(1000*pdflmax); // maximum for random number
                            randprob=((double)(rand()%randmax))/1000; // random number for probability

                            if ( randprob < PDFL[randangl] ) {
                                PAR_MPAR2[0] = (double)(PDFL_O[randangl]) / 1000; // maximum travel distance
                                ctrl_mpar1 = 0;
                            }
                        }

                        else if ( PAR_CTRLCDF == 3 ) {

                            randmax=(int)(1000*pdfomax) * (int)(1000*pdflmax); // maximum for random number
                            randprob=((double)(rand()%randmax))/1000000; // random number for probability

                            if ( randprob < PDFO[randango] * PDFL[randangl] ) {
                                PAR_MPAR1[0] = (double)(PDFO_O[randango]) / 1000 * 180 / CONST_PI; // angle of reach
                                PAR_MPAR2[0] = (double)(PDFL_O[randangl]) / 1000; // maximum travel distance
                                ctrl_mpar1 = 0;
                            }
                        }
                    }

                    if ( PAR_SWALK == 0 || PAR_NMODELS == 1 ) {
                        // if separate routing for each model does not apply, or if only one model is used:

                        model_min = 0; // id of first model to be applied 
                        model_max = PAR_NMODELS; // id of last model to be applied
                    }
                    else { // if separate routing applies, selecting model randomly:

                        randmax=PAR_NMODELS; // maximum for random number
                        randnum=(double)(rand()%randmax); // random number

                        model_min = (int)randnum; // id of first model to be applied 
                        model_max = (int)randnum + 1; // id of last model to be applied
                    }

                    while ( ctrlall > 0 && ctrl_valid == 1 && ctrl_impbar <= 0 && ctr_run < 10000 ) { // routing random walk:


                    // DETERMINING TARGET PIXEL -----------------------------------------------


                        ctr_run += 1; // updating number of step

                        xf_run[ctr_run] = xf; yf_run[ctr_run] = yf; // storing coordinates in array

                        if ( PAR_CTRLFLOWDIR == 1 ) { // if flow direction is defined:

                            #ifdef WITHGRASS

                                Segment_get(&seg_flowdir, &fdir, xf, yf); // flow direction

                            #else

                                fdir = seg_flowdir[xf][yf]; // flow direction

                            #endif

                            if (fdir == 4 || fdir == 5 || fdir == 6 ) xf += 1; // updating coordinates
                            else if (fdir == 8 || fdir == 1 || fdir == 2 ) xf -= 1;
                            if (fdir == 6 || fdir == 7 || fdir == 8 ) yf -= 1;
                            else if (fdir == 2 || fdir == 3 || fdir == 4 ) yf += 1;
                        }
                        else { // if flow direction is not defined:

                            ctr_run_test = ctr_run - 1; // identifying control point:
                            if ( PAR_SCTRL[model_min] >= s_mov[ctr_run-1] ) ctr_run_test = 0;
                            else {
                                while ( s_mov[ctr_run-1] - s_mov[ctr_run_test] < PAR_SCTRL[model_min] && ctr_run_test > 0 )
                                    ctr_run_test -= 1;
                            }

                            dist_ctrl_prevapp = pow ( pow ( xf * sico.CSZ - xf_run[ctr_run_test] * sico.CSZ, 2 )
                                + pow ( yf * sico.CSZ - yf_run[ctr_run_test] * sico.CSZ, 2 ), 0.5 ); // distance to control point

                            if ( PAR_SEG[model_min] + s_mov[0] >= s_mov[ctr_run-1] || ctr_run == 1 ) {

                                ctr_run_seg = 0;
                                elev_seg_prev = elev_release;
                                s_mov_segstart00 = 0;
                                xf_segstart00 = x;
                                yf_segstart00 = y;
                            }
                            else {
                                ctr_run_seg = ctr_run - 1; // identifying segment point:

                                while ( pow(pow(sico.CSZ*xf_run[ctr_run]-sico.CSZ*xf_run[ctr_run_seg],2)
                                    + pow(sico.CSZ*yf_run[ctr_run]-sico.CSZ*yf_run[ctr_run_seg],2),0.5)
                                    < PAR_SEG[model_min] && ctr_run_seg > 1 ) ctr_run_seg -= 1;

                                s_mov_segstart00 = s_mov[ctr_run_seg]; // travel distance at segment point
                                xf_segstart00 = xf_run[ctr_run_seg];
                                yf_segstart00 = yf_run[ctr_run_seg];

                                #ifdef WITHGRASS

                                    Segment_get(&seg_elev, &elev_seg_prev, xf_run[ctr_run_seg], yf_run[ctr_run_seg]);
                                        // elevation of segment point

                                #else

                                    elev_seg_prev = seg_elev[xf_run[ctr_run_seg]][yf_run[ctr_run_seg]]; // elevation of segment point

                                #endif

                            }

                            #ifdef WITHGRASS

                                Segment_get(&seg_elev, &elev_ctrl_prev, xf_run[ctr_run_test], yf_run[ctr_run_test]);
                                    // elevation of control point

                            #else

                                elev_ctrl_prev = seg_elev[xf_run[ctr_run_test]][yf_run[ctr_run_test]]; // elevation of control point

                            #endif

                            xadj[1]=xf-1; yadj[1]=yf+1; // coordinates of adjacent cells
                            xadj[2]=xf-1; yadj[2]=yf;
                            xadj[3]=xf-1; yadj[3]=yf-1;
                            xadj[4]=xf; yadj[4]=yf-1;
                            xadj[5]=xf+1; yadj[5]=yf-1;
                            xadj[6]=xf+1; yadj[6]=yf;
                            xadj[7]=xf+1; yadj[7]=yf+1;
                            xadj[8]=xf; yadj[8]=yf+1;
                            whtslope[0]=0; // initializing weighted slope

                            #ifdef WITHGRASS

                                Segment_get(&seg_elev, &elev, xf, yf); // elevation

                            #else

                                elev = seg_elev[xf][yf]; // elevation

                            #endif

                            for (fdir=1; fdir<9; fdir++) { // starting loop over all neighbouring cells:

                                if ( fdir == 1 || fdir == 3 || fdir == 5 || fdir == 7 ) deltas = sico.CSZ * pow (2, 0.5);
                                else deltas = sico.CSZ; // updating distance from release

                                #ifdef WITHGRASS

                                    Segment_get(&seg_elev, &elev_adj, xadj[fdir], yadj[fdir]); // elevation of neighbour pixel
                                    Segment_get(&seg_aoi, &ctrl_aoi, xadj[fdir], yadj[fdir]); // control for aoi of neighbour pixel

                                #else

                                    elev_adj = seg_elev[xadj[fdir]][yadj[fdir]]; // elevation of neighbour pixel
                                    ctrl_aoi = seg_aoi[xadj[fdir]][yadj[fdir]]; // control for aoi of neighbour pixel

                                #endif

                                dist_ctrl_app = pow ( pow ( xadj[fdir] * sico.CSZ - xf_run[ctr_run_test] * sico.CSZ, 2 )
                                    + pow ( yadj[fdir] * sico.CSZ - yf_run[ctr_run_test] * sico.CSZ, 2 ), 0.5 );
                                    // distance of neighbour pixel to control point

                                if ( elev_adj == sico.UDM && ctrl_aoi <= 0 ) ctrl_valid2 = 0;
                                    // checking whether elevation information is available

                                if ( dist_ctrl_app > dist_ctrl_prevapp
                                    && z_movmax - ( elev_release - elev_adj ) <= PAR_RUNUPMAX[model_min] ) {
                                    // if outflow to this pixel is possible:

                                    if ( fdir == fdir_prev ) direction_factor = 100 * pow( PAR_FDIR[model_min], 2 );
                                    else if ( fdir == fdir_prev - 1 || fdir == fdir_prev + 1
                                        || ( fdir == 1 && fdir_prev == 8) || ( fdir == 8 && fdir_prev == 1) )
                                        direction_factor = 100 * PAR_FDIR[model_min];
                                    else direction_factor = 100; // weighting for maintenance of flow direction

                                    if (( elev - elev_adj ) / deltas <= 2.75 ) // weighting for slope:
                                        whtslope[fdir] = whtslope[fdir-1] + direction_factor
                                            * exp(( elev - elev_adj ) / deltas * PAR_FSLOPE[model_min] );
                                    else whtslope[fdir] = whtslope[fdir-1] + direction_factor * exp( 2.75 * PAR_FSLOPE[model_min] );
                                        // cumulative weighted slope

                                    /*if ( xadj[fdir] != xf_run[ctr_run_test] )
                                        betax = atan ( elev_ctrl_prev - elev_adj ) / (( xadj[fdir] - xf_run[ctr_run_test] ) * sico.CSZ );
                                    else betax = 0;
                                    if ( yadj[fdir] != yf_run[ctr_run_test] )
                                        betay = atan ( elev_ctrl_prev - elev_adj ) / (( yadj[fdir] - yf_run[ctr_run_test] ) * sico.CSZ );
                                    else betay = 0; // x and y slopes

				    alphatestd[fdir] = CONST_PI / 180 * falpha( betax, betay ); // previous flow direction

                                    if ( alphaprev == sico.UDM ) alphaprev = alphatestd[fdir];
                                    alphadiff1 = fabs( alphatestd[fdir] - alphaprev );
                                    alphadiff2 = fabs( alphatestd[fdir] - alphaprev - 2 * CONST_PI );
                                    if ( alphadiff1 < alphadiff2 ) alphadiff = alphadiff1; else alphadiff = alphadiff2;

                                    if ( alphadiff < CONST_PI / 2 ) {

                                        if (( elev_ctrl_prev - elev_adj ) / dist_ctrl_app <= 2.75 ) // weighting
                                            whtslope[fdir] = whtslope[fdir-1] + 1000 * exp( - PAR_FDIR[model_min] * tan( alphadiff ))
                                                * exp(( elev_ctrl_prev - elev_adj ) / dist_ctrl_app * PAR_FSLOPE[model_min] );
                                        else whtslope[fdir] = whtslope[fdir-1] + 1000 * exp( - PAR_FDIR[model_min] * tan( alphadiff ))
                                                * exp( 2.75 * PAR_FSLOPE[model_min] );
                                    }
                                    else whtslope[fdir] = whtslope[fdir-1]; */
                                }
                                else whtslope[fdir] = whtslope[fdir-1];
                                    // if outflow to this pixel is not possible, not increasing cumulative weighted slope
                                if ( fdir == 8 && whtslope[fdir] > 0 && whtslope[fdir] < 0.5 ) whtslope[fdir] = 0.5;
                                    // ensuring minimum of weighted slope
                            }

                            if ( whtslope[8] > 0 ) { // randomly determining target pixel:

                                randmax=(int)(whtslope[8]+0.5); // maximum for random number
                                randnum=(double)(rand()%randmax); // random number
                                if (randnum < 0.0001) randnum=0.0001;

                                for (l=1; l<9; l++) { // starting loop over all downslope directions
                                    if (randnum <= whtslope[l]) { // if random number matches direction:
                                        fdir=l; // direction
                                        //alphaprev = alphatestd[l];
                                        break; // leaving loop
                                    }
                                }
                            }
                            else ctrl_valid = 0; // if no outflow is possible, setting validity to negative
                            fdir_prev = fdir;

                            if (fdir == 5 || fdir == 6 || fdir == 7 ) xf += 1; // updating coordinates
                            else if (fdir == 1 || fdir == 2 || fdir == 3 ) xf -= 1;
                            if (fdir == 3 || fdir == 4 || fdir == 5 ) yf -= 1;
                            else if (fdir == 1 || fdir == 7 || fdir == 8 ) yf += 1;
                        }

                        #ifdef WITHGRASS

                            Segment_get(&seg_elev, &elev_f, xf, yf); // elevation of target pixel
                            Segment_get(&seg_aoi, &aoi_f, xf, yf); // control for aoi of target pixel

                        #else

                            elev_f = seg_elev[xf][yf]; // elevation of target pixel
                            aoi_f = seg_aoi[xf][yf]; // control for aoi of target pixel

                        #endif

                        if ( ctrl_valid == 1 ) { // updating distance from release coordinates:

                            s_mov_seg = pow ( pow ( xf * sico.CSZ - xseg * sico.CSZ, 2 ) + pow ( yf * sico.CSZ - yseg * sico.CSZ, 2 ), 0.5 );
                            s_mov[ctr_run] = s_mov_segstart + s_mov_seg;

                            if ( s_mov_seg >= PAR_SEG[model_min] ) {

                                xseg = xf;
                                yseg = yf;
                                s_mov_segstart = s_mov[ctr_run];
                                s_mov_segstart0 = s_mov[ctr_run-1];
                                elev_segstart0 = elev;
                            }

                            z_movprev = z_mov;
                            z_mov = elev_release-elev_f; // updating total loss of elevation
                            if ( z_mov > z_movmax ) z_movmax = z_mov; // updating maximum loss of elevation
                        }

                        if ( xf < 2 || xf > (sico.M-3) || yf < 2 || yf > (sico.N-3) || aoi_f <= 0 || ctrl_valid2 == 0 ) {
                            // if movement would leave valid zone:

                            ctr_failalloc+=1; // updating counter for unsuccessful routing attempts
                            for ( model = model_min; model < model_max; model++ ) {
                                // setting maximum travel distance and angle of reach path to undefined:
                                s_movmax[model] = sico.UDM;
                                omega_stop[model] = sico.UDP;
                            }
                            break; // leaving loop
                        }

                        if ( ctrl_valid == 0 ) break; // leaving loop if no more proceeding is possible

                        if ( PAR_CTRLIMPOBJ == 1 && PAR_CTRLOBJSCORES == 1 ) { // testing whether random walk hits a barrier:

                            #ifdef WITHGRASS

                                Segment_get(&seg_impobj, &ctrl_impobj, xf, yf); // impact objects

                            #else

                                ctrl_impobj = seg_impobj[xf][yf]; // impact objects

                            #endif

                            for ( nimpobj = 0; nimpobj < PAR_NIMPOBJ; nimpobj++ ) { // searching for correct object id:
                                if ( IMP_NAME[nimpobj] == ctrl_impobj && IMP_TYPE[nimpobj] == 2 ) {
                                    ctrl_impbar = 1; // if impact object is defined as a barrier, setting control to positive
                                    break; // leaving loop
                                }
                            }
                        }


                    // APPLYING MODELS --------------------------------------------------------


                        #ifdef WITHGRASS

                            if ( PAR_BACK == 1  ) Segment_get(&seg_imparea, &imparea, xf, yf); // impact area map

                        #else

                            if ( PAR_BACK == 1  ) imparea = seg_imparea[xf][yf]; // impact area map

                        #endif

                        omega[ctr_run] = 180 / CONST_PI * atan ( z_mov / s_mov[ctr_run] ); // average angle of path

                        for ( model = model_min; model < model_max; model++ ) { // loop over all models:

                            if ( ctrl_M[model] == 1 ) { // if movement is still active:

                                if ( PAR_MTYPE[model] == 1 ) // model equations:
                                    omega_test[model] = PAR_MPAR1[model];
                                else if ( PAR_MTYPE[model] == 2 )
                                    omega_test[model] =
                                        180 / CONST_PI * pow( 10, atan( PAR_MPAR1[model] * log10( casev ) + PAR_MPAR2[model] ));
                                else if ( PAR_MTYPE[model] == 3 )
                                    smov_test = PAR_MPAR1[model] * pow( casev, PAR_MPAR2[model] ) * pow( z_mov, PAR_MPAR3[model] );
                                else if ( PAR_MTYPE[model] == 4 )
                                    omega_test[model] = PAR_MPAR1[model] * pow( caseq, PAR_MPAR2[model] );
                                else if ( PAR_MTYPE[model] == 0 && ctr_run > 0 ) {

                                    if ( ctr_run == 1 ) omegatop = 180 / CONST_PI * atan ( z_mov / s_mov[1] );
                                    omegatrans = 180 / CONST_PI * atan (( z_mov - z_movprev ) / ( s_mov[ctr_run] - s_mov[ctr_run-1] ));
                                    if ( omegatrans >= 10 ) omegaup = omega[ctr_run];

                                    //omega_test[model] = ( 0.62 - 0.28 * z_mov * 0.0006 ) * omegaup + ( 19 * z_mov * 0.0006 - 2.3 ) + 0.12 * omegatop;
                                    omega_test[model] = PAR_MPAR1[model] * omegaup + PAR_MPAR2[model];
                                }
                                else if ( PAR_MTYPE[model] == 5 ) { // two-parameter friction model:

                                    xalpha[1] = atan(( elev_seg_prev - elev_f ) / ( pow(pow(sico.CSZ*xf_run[ctr_run]-sico.CSZ*xf_segstart00,2)
                                        + pow(sico.CSZ*yf_run[ctr_run]-sico.CSZ*yf_segstart00,2),0.5) )); // slope

                                    if ( ctr_run == 1 || xalpha[1] > xalpha[0] ) xdalpha = 0;
                                    else xdalpha = xalpha[0] - xalpha[1]; // curvature

                                    xv2 = xv1 * cos( xdalpha ); // corrected flow velocity

                                    xzeta = 9.81 * ( sin( xalpha[1] ) - pow( 10, PAR_MPAR1[model] ) * cos( xalpha[1] ));
                                    xeta = - 2 * ( deltas ) / cos( xalpha[1] ) / pow ( 10, PAR_MPAR2[model]);
                                    if ( xzeta * pow ( 10, PAR_MPAR2[model] ) * ( 1 - exp( xeta )) + pow( xv2, 2 )
                                        * exp( xeta ) * cos( xdalpha ) > 0 )
                                        xv = pow( xzeta * pow ( 10, PAR_MPAR2[model] ) * ( 1 - exp( xeta ))
                                            + pow( xv2, 2 ) * exp( xeta ) * cos( xdalpha ), 0.5 ); // updating flow velocity
                                    else xv = 0;

                                    xalpha[0] = xalpha[1]; // storing slope for next step
                                    xv1 = xv;
                                }

                                if ( PAR_NRUNS0 == sico.UDM ) {
                                    if ( PAR_CTRLCDF == 1 || PAR_CTRLCDF == 3 ) omega_test[model] = PAR_MPAR1[model]; else omega_test[model] = 0;
                                    if ( PAR_CTRLCDF == 2 || PAR_CTRLCDF == 3 ) smov_test = PAR_MPAR2[model]; else smov_test = pow(10,9);
                                }


                            // UPDATING RESULTS -------------------------------------------------------


                                if (((( PAR_MTYPE[model] == 5 && xv > 0 ) || ( PAR_MTYPE[model] == 3 && s_mov[ctr_run] <= smov_test )
                                    || ( PAR_NRUNS0 == sico.UDM && ( s_mov[ctr_run] <= smov_test && omega[ctr_run] >= omega_test[model] )) 
                                    || (( PAR_MTYPE[model] == 1 || PAR_MTYPE[model] == 2 || PAR_MTYPE[model] == 4 || PAR_MTYPE[model] == 0 )
                                    && omega[ctr_run] >= omega_test[model]) && PAR_BACK == 0 && PAR_NRUNS0 != sico.UDM )
                                    || ( PAR_BACK > 0 && imparea >= casename3 )) || s_mov[ctr_run] < PAR_SMIN ) {
                                    // if model is still active: !!!BACK-CALC WILL ONLY STOP WHEN REACHING LOWER ID
                                }
                                else ctrl_M[model] = 0; // else, terminating motion
                            }
                        }

                        ctrlall = 0; for ( model = model_min; model < model_max; model++ ) ctrlall += ctrl_M[model];
                            // number of active models

                        if ( ctrlall > 0 ) { // if at least one model is still active:

                            ctrl_jimpact = 1; // initializing control for newly affected pixel
                            for ( iimpact = 0; iimpact <= jimpact; iimpact++ ) { // loop over all cells already impacted:

                                if ( xf == ximpact[iimpact] && yf == yimpact[iimpact] ) { // testing if cell was already impacted:
                                    ctrl_jimpact = 0; // setting control for newly affected pixel to negative
                                    kimpact = iimpact; // updating index of current pixel
                                }
                            }
                            if ( ctrl_jimpact == 1 ) { // if cell was not yet impacted:
                                jimpact += 1; // increasing counter
                                kimpact = jimpact; // updating index of current pixel
                                ximpact[kimpact] = xf; // storing coordinates
                                yimpact[kimpact] = yf;
                                for ( model = 0; model < PAR_NMODELS; model++ ) iM[model][kimpact] = 0; // resetting impact frequency
                                s_mov_out[kimpact] = 0; z_mov_out[kimpact] = sico.UDM; omega_out[kimpact] = sico.UDM;
                                if ( PAR_CTRLVEL == 1 ) xxv[kimpact] = sico.UDM;
                                if ( PAR_PROB == 1 ) { impprobr[kimpact] = 0; impprob[kimpact] = 0; impquant[kimpact] = 0; }
                                if ( PAR_SCORE == 1 ) his[kimpact] = sico.UDM; // resetting array elements
                            }

                            if ( z_movmax - z_mov > runupmax ) runupmax = z_movmax - z_mov; // updating maximum runup height

                            if ( s_mov_out[kimpact] == 0 || s_mov[ctr_run] < s_mov_out[kimpact] ) s_mov_out[kimpact] = s_mov[ctr_run];
                                // updating travel distance for output
                            if ( z_mov_out[kimpact] == 0 ) z_mov_out[kimpact] = z_mov; // updating loss of elevation
                            if ( omega[ctr_run] > omega_out[kimpact] ) omega_out[kimpact] = omega[ctr_run];
                                // updating average angle of path

                            if ( PAR_CTRLIMPOBJ == 1 ) {
                                // updating distances and average angle of path related to impact objects:

                                #ifdef WITHGRASS

                                    Segment_get(&seg_impobj, &ctrl_impobj, xf, yf); // impact objects

                                #else

                                    ctrl_impobj = seg_impobj[xf][yf]; // impact objects

                                #endif

                                if ( ctrl_impobj > 0 && PAR_CONN == 1 ) lconn += ctrlall;

                                if ( ctrl_impobj > 0 && ctrl_impact == 0 ) {

                                    if ( s_mov[ctr_run] < smovi ) smovi = s_mov[ctr_run];
                                    if ( z_mov < zmovi ) zmovi = z_mov;
                                    if ( omega[ctr_run] > omegai ) omegai = omega[ctr_run];
                                    ctrl_impact = 1;
                                }
                            }

                            if ( PAR_CTRLVEL == 1 && xv1 > xxv[kimpact] ) xxv[kimpact] = xv1; // updating velocity

                            impquant[kimpact] = 1; // updating number of cases possibly affecting the pixel
                            for ( model = model_min; model < model_max; model++ ) {
                                if ( ctrl_M[model] == 1 ) iM[model][kimpact] += 1; // updating impact frequency
                            }

                            if ( PAR_PROB == 1 ) { // for probabilities:

                                if ( PAR_CTRLCDF == 1 || PAR_CTRLCDF == 3 ) {

                                    if ( tan( omega[ctr_run] * CONST_PI / 180 ) < 2 )
                                        impprobr_testo = CDFO[(int)(tan(omega[ctr_run]*CONST_PI/180)*1000+0.5)]; // impact probability
                                    else impprobr_testo = 1;
                                }
                                else impprobr_testo = 1;

                                if ( PAR_CTRLCDF == 2 || PAR_CTRLCDF == 3 ) {

                                    if ( s_mov[ctr_run] <= ((double)(CDFL_O[2000])/1000))
                                        impprobr_testl = 1-CDFL[(int)(2000*s_mov[ctr_run]/((float)(CDFL_O[2000])/1000)+0.5)]; // impact probability
                                    else impprobr_testl = 0;
                                }
                                else impprobr_testl = 1;

                                impprobr_test = impprobr_testo * impprobr_testl;
                                if ( impprobr_test > impprobr[kimpact] ) impprobr[kimpact] = impprobr_test; // updating impact probability
                            }

                            if ( PAR_SCORE == 1 ) { // for scores:

                                if ( ctr_rscore == sico.UDM ) his_test = ctrlall; // impact hazard indication score
                                else his_test = (int)( pow( pow( (double)ctr_rscore, 1.5 )
                                    * pow ( (double)iis[iimpact], 1.5 ), (double)1 / (double)3 ));
                                if ( his_test > his[kimpact] ) his[kimpact] = his_test; // updating impact hazard indication score
                            }
                        }
                    } // stopping random walk
                } // stopping series of random walks for start pixel/coordinates

                if ( PAR_CTRLIMPOBJ == 1 ) { // updating distances and average angle of path with regard to impact objects:

                    #ifdef WITHGRASS

                        Segment_put(&seg_smovi, &smovi, x, y);
                        Segment_put(&seg_zmovi, &zmovi, x, y);
                        Segment_put(&seg_omegai, &omegai, x, y);

                    #else

                        seg_smovi = smovi;
                        seg_zmovi = zmovi;
                        seg_omegai = omegai;

                    #endif

                }

                if ( PAR_CONN == 1 ) { // updating connectivity:

                    #ifdef WITHGRASS

                        Segment_put(&seg_conn, &lconn, x, y);

                    #else

                        seg_conn[x][y] = lconn;

                    #endif

                    if ( PAR_NCASES > 0 ) fprintf(f_conn, "%i\t%.0f\t%.0f\t%i\n", casename, CASE_XR[caseid], CASE_YR[caseid], lconn);
                }
            }

            if ( PAR_CONN == 1 ) {

                #ifdef WITHGRASS

                    Segment_flush(&seg_conn);

                #endif

            }
        }}


    // UPDATING OUTPUT WITH RESULTS FOR THE CONSIDERED CASE -----------------------------------


        for ( iimpact = 0; iimpact <= jimpact; iimpact++ ) { // loop over all impacted pixels:

            iis[iimpact] = 0; // impact indicator score:

            for ( model = 0; model < PAR_NMODELS; model++ ) {

                if ( iM[model][iimpact] > 0 ) iis[iimpact] += 1; // updating impact indicator score
                if ( s_mov_out[iimpact] > s_movmax[model] && iM[model][iimpact] > 0 ) {
                    s_movmax[model] = s_mov_out[iimpact]; // updating maximum travel distance
                    omega_stop[model] = omega_out[iimpact]; // updating average angle of flow path
                }
            }

            #ifdef WITHGRASS

                Segment_get(&seg_iis, &iisold, ximpact[iimpact], yimpact[iimpact]); // reading old impact indicator score
                if ( iis[iimpact] > iisold ) { // if score exceeds the old value:
                    Segment_put(&seg_iis, &iis[iimpact], ximpact[iimpact], yimpact[iimpact]); // updating score
                    Segment_flush(&seg_iis);
                }

                for ( model = 0; model < PAR_NMODELS; model++ ) { // loop over all models:

                    Segment_get(&seg_iM[model], &iMold, ximpact[iimpact], yimpact[iimpact]); // old impact frequency
                    iMnew = iMold + iM[model][iimpact]; // updating impact frequency for the considered pixel
                    Segment_put(&seg_iM[model], &iMnew, ximpact[iimpact], yimpact[iimpact]);
                    Segment_flush(&seg_iM[model]);
                }

                if ( PAR_MULT == 0 ) { // for single model run:

                    Segment_get(&seg_smov, &smov_outold, ximpact[iimpact], yimpact[iimpact]); // old travel distance
                    if ( smov_outold == 0 || s_mov_out[iimpact] < smov_outold ) { // if distance is smaller than the old value for the pixel:
                        Segment_put(&seg_smov, &s_mov_out[iimpact], ximpact[iimpact], yimpact[iimpact]); // updating distance
                        Segment_flush(&seg_smov);
                    }

                    Segment_get(&seg_zmov, &zmov_outold, ximpact[iimpact], yimpact[iimpact]); // old loss of elevation
                    if ( zmov_outold == 0 || z_mov_out[iimpact] < zmov_outold ) { // if loss is smaller than the old value for the pixel:
                        Segment_put(&seg_zmov, &z_mov_out[iimpact], ximpact[iimpact], yimpact[iimpact]); // updating distance
                        Segment_flush(&seg_zmov);
                    }

                    Segment_get(&seg_omega, &omega_outold, ximpact[iimpact], yimpact[iimpact]); // old average angle of path
                    if ( omega_out[iimpact] > omega_outold ) { // if angle exceeds the old value for the considered pixel:
                        Segment_put(&seg_omega, &omega_out[iimpact], ximpact[iimpact], yimpact[iimpact]); // updating angle
                        Segment_flush(&seg_omega);
                    }

                    if ( PAR_CTRLVEL == 1 ) {

                        Segment_get(&seg_xv, &xv, ximpact[iimpact], yimpact[iimpact]); // old velocity
                        if ( xxv[iimpact] > xv ) { // if velocity exceeds the old value for the considered pixel:
                            Segment_put(&seg_xv, &xxv[iimpact], ximpact[iimpact], yimpact[iimpact]); // updating velocity
                            Segment_flush(&seg_xv);
                        }
                    }

                    if ( PAR_SCORE == 1 ) { // hazard indicator score:

                        Segment_get(&seg_his, &hisold, ximpact[iimpact], yimpact[iimpact]); // old impact hazard indicator score
                        if ( his[iimpact] > hisold ) { // if score exceeds the old value for the considered pixel:
                            Segment_put(&seg_his, &his[iimpact], ximpact[iimpact], yimpact[iimpact]); // updating score
                            Segment_flush(&seg_his);
                        }
                    }

                    if ( PAR_PROB == 1 ) { // average release probability of possible release zone:

                        Segment_get(&seg_relprobr, &relprobrold, ximpact[iimpact], yimpact[iimpact]); // old sum of release probabilities
                        relprobrnew = relprobrold + prob; // updating sum
                        if ( relprobrnew > 0 ) {
                            Segment_put(&seg_relprobr, &relprobrnew, ximpact[iimpact], yimpact[iimpact]);
                            Segment_flush(&seg_relprobr);
                        }
                    }
                }

                Segment_get(&seg_impquant, &impquantold, ximpact[iimpact], yimpact[iimpact]);
                    // old number of cases possibly affecting the pixel
                impquantnew = impquantold + impquant[iimpact]; // updating area
                if ( impquantnew > 0 ) {
                    Segment_put(&seg_impquant, &impquantnew, ximpact[iimpact], yimpact[iimpact]);
                    Segment_flush(&seg_impquant);
                }

                if ( PAR_PROB == 1 ) { // for probabilities:

                    Segment_get(&seg_impprobr, &impprobrold, ximpact[iimpact], yimpact[iimpact]); // old (sum of) impact probability

                    if ( PAR_AFLAG == 0 ) {
                        if ( impprobr[iimpact] > impprobrold ) impprobrnew = impprobr[iimpact]; // updating impact probability:
                        else impprobrnew = 0;
                    }
                    else impprobrnew = impprobrold + impprobr[iimpact]; // updating sum of impact probability

                    if ( impprobrnew > 0 ) {
                        Segment_put(&seg_impprobr, &impprobrnew, ximpact[iimpact], yimpact[iimpact]);
                        Segment_flush(&seg_impprobr);
                    }
                }

            #else

                iisold = seg_iis[ximpact[iimpact]][yimpact[iimpact]]; // old impact indicator score
                if ( iis[iimpact] > iisold ) seg_iis[ximpact[iimpact]][yimpact[iimpact]] = iis[iimpact]; // updating score

                for ( model = 0; model < PAR_NMODELS; model++ ) { // loop over all models:

                    iMold = seg_iM[model][ximpact[iimpact]][yimpact[iimpact]]; // old impact frequency
                    iMnew = iMold + iM[model][iimpact]; // updating impact frequency for the considered pixel
                    seg_iM[model][ximpact[iimpact]][yimpact[iimpact]] = iMnew;
                }

                if ( PAR_MULT == 0 ) { // for single model run:

                    smov_outold = seg_smov[ximpact[iimpact]][yimpact[iimpact]]; // old travel distance
                    if ( smov_outold == 0 || s_mov_out[iimpact] < smov_outold ) // if distance is smaller than the old value for the pixel:
                        seg_smov[ximpact[iimpact]][yimpact[iimpact]] = s_mov_out[iimpact]; // updating distance

                    zmov_outold = seg_zmov[ximpact[iimpact]][yimpact[iimpact]]; // old loss of elevation
                    if ( zmov_outold == 0 || z_mov_out[iimpact] < zmov_outold ) // if loss is smaller than the old value for the pixel:
                        seg_zmov[ximpact[iimpact]][yimpact[iimpact]] = z_mov_out[iimpact]; // updating distance

                    omega_outold = seg_omega[ximpact[iimpact]][yimpact[iimpact]]; // old average angle of path
                    if ( omega_out[iimpact] > omega_outold ) // if angle exceeds the old value for the considered pixel:
                        seg_omega[ximpact[iimpact]][yimpact[iimpact]] = omega_out[iimpact]; // updating angle

                    if ( PAR_CTRLVEL == 1 ) {

                        xv = seg_xv[ximpact[iimpact]][yimpact[iimpact]]; // old velocity
                        if ( xxv[iimpact] > xv ) // if velocity exceeds the old value for the considered pixel:
                            seg_xv[ximpact[iimpact]][yimpact[iimpact]] = xxv[iimpact]; // updating velocity
                    }

                    if ( PAR_SCORE == 1 ) { // hazard indicator score:

                        hisold = seg_his[ximpact[iimpact]][yimpact[iimpact]]; // old impact hazard indicator score
                        if ( his[iimpact] > hisold ) // if score exceeds the old value for the considered pixel:
                            seg_his[ximpact[iimpact]][yimpact[iimpact]] = his[iimpact]; // updating score
                    }

                    if ( PAR_PROB == 1 ) { // average release probability of possible release zone:

                        relprobrold = seg_relprobr[ximpact[iimpact]][yimpact[iimpact]]; // old sum of release probabilities
                        relprobrnew = relprobrold + prob; // updating sum
                        if ( relprobrnew > 0 ) seg_relprobr[ximpact[iimpact]][yimpact[iimpact]] = relprobrnew;
                    }
                }

                impquantold = seg_impquant[ximpact[iimpact]][yimpact[iimpact]]; // old number of cases possibly affecting the pixel
                impquantnew = impquantold + impquant[iimpact]; // updating area
                if ( impquantnew > 0 ) seg_impquant[ximpact[iimpact]][yimpact[iimpact]] =impquantnew;

                if ( PAR_PROB == 1 ) { // for probabilities:

                    impprobrold = seg_impprobr[ximpact[iimpact]][yimpact[iimpact]]; // old (sum of) impact probability

                    if ( PAR_AFLAG == 0 ) {
                        if ( impprobr[iimpact] > impprobrold ) impprobrnew = impprobr[iimpact]; // updating impact probability:
                        else impprobrnew = 0;
                    }
                    else impprobrnew = impprobrold + impprobr[iimpact]; // updating sum of impact probability
                    if ( impprobrnew > 0 ) seg_impprobr[ximpact[iimpact]][yimpact[iimpact]] = impprobrnew;
                }

            #endif

        }

        for ( model = 0; model < PAR_NMODELS; model++ ) {
            if ( omega_stop[model] == sico.UDP ) omega_stop[model] = sico.UDM;
                // if necessary, setting angle of reach to no data
        }

        if ( jimpact > 0 ) { // printing summary to file and display:

            fprintf(f_summary, "%i", casename);
            printf("   %i\tOK   ", casename);

            for ( model = 0; model < PAR_NMODELS; model++ ) {
                fprintf(f_summary, "\t%.0f\t%.2f", s_movmax[model], omega_stop[model]);
                printf("\t%.0f\t%.2f", s_movmax[model], omega_stop[model]);
            }
            fprintf(f_summary, "\t%.1f\t%.0f\n", runupmax, jimpact * pow( sico.CSZ, 2 ));
            printf("\t%.1f\t%.0f\n", runupmax, jimpact * pow( sico.CSZ, 2 ));

            if ( PAR_BACK == 1 && s_movmax[0] >= PAR_SMIN ) { // for back-calculation:

                rid += 1; // updating case id for back-calculation output
                fprintf(f_backfile, "%i\t%i\t%.1f\t%.3f\n", rid, casename3, s_movmax[0], omega_stop[0] );
                    // writing detail output to file
            }
        }

    } // stopping loop over all cases

    if ( PAR_CTRLIMPOBJ == 1 ) {

        #ifdef WITHGRASS

            Segment_flush(&seg_smovi);
            Segment_flush(&seg_zmovi);
            Segment_flush(&seg_omegai);

        #endif
    }


// COMPUTING RISK INDICATOR SCORE -------------------------------------------------------------


    nimpact_all = 0; // initializing total number of affected pixels

    for (x=0; x<sico.M; x++) { // loop over all pixels:
        for (y=0; y<sico.N; y++) {

            #ifdef WITHGRASS

                Segment_get(&seg_aoi, &ctrl_aoi, x, y); // control for area of interest
                Segment_get(&seg_iis, &iisold, x, y); // impact indicator score

            #else

                ctrl_aoi = seg_aoi[x][y]; // control for area of interest
                iisold = seg_iis[x][y]; // impact indicator score

            #endif

            if ( iisold > 0 ) nimpact_all += 1; // updating total number of affected pixels

            if ( PAR_CTRLIMPOBJ == 1 && ctrl_aoi == 1 ) {

                #ifdef WITHGRASS

                    if ( PAR_SCORE == 1 ) Segment_get(&seg_his, &his_test, x, y); // impact hazard indicator score
                    Segment_get(&seg_impobj, &ctrl_impobj, x, y); // impact objects

                #else

                    if ( PAR_SCORE == 1 ) his_test = seg_his[x][y]; // impact hazard indicator score
                    ctrl_impobj = seg_impobj[x][y]; // impact objects

                #endif

                if ( PAR_SCORE == 1 && ctrl_impobj > 0 && his_test > 0 ) {

                    if ( PAR_CTRLOBJSCORES == 1 ) { // if object scores are available:

                        for ( nimpobj = 0; nimpobj < PAR_NIMPOBJ; nimpobj++ ) // searching for correct impact object id
                            if ( IMP_NAME[nimpobj] == ctrl_impobj ) break; // leaving loop when id has been found

                        ris = (int)( pow( pow( (double)his_test, 1.5 ) * pow ( (double)IMP_S[nimpobj], 1.5 ), (double)1 / (double)3 ));
                            // risk indicator score

                        if ( ris > ris_obj[nimpobj] ) ris_obj[nimpobj] = ris; // updating maximum risk indicator score of object
                    }
                    else ris = his_test;

                    #ifdef WITHGRASS

                        Segment_put(&seg_ris_impobj, &ris, x, y);

                    #else

                        seg_ris_impobj[x][y] = ris;

                    #endif
                }
            }
        }
    }

    #ifdef WITHGRASS

        if ( PAR_CTRLIMPOBJ == 1 ) Segment_flush(&seg_ris_impobj);

    #endif

    if ( PAR_CTRLIMPOBJ == 1 && PAR_CTRLOBJSCORES == 1 ) {
        for ( nimpobj = 0; nimpobj < PAR_NIMPOBJ; nimpobj++ ) // updating risk indicator score file:
            fprintf(f_impobj, "%i %i %i %i", IMP_NAME[nimpobj], IMP_TYPE[nimpobj], IMP_S[nimpobj], ris_obj[nimpobj]);
        fclose(f_impobj);
    }

    if ( PAR_MULT == 1 ) fprintf(f_nout, "%i\n%.1f\n%.3f\n1\n", nimpact_all, s_movmax[0], omega_stop[0] );
        // updating file with information on model run


// FINALIZING PROBABILITIES -------------------------------------------------------------------


    if ( PAR_PROB == 1 ) {

        for (x=0; x<sico.M; x++) { // loop over all pixels:
            for (y=0; y<sico.N; y++) {

                #ifdef WITHGRASS

                    Segment_get(&seg_aoi, &ctrl_aoi, x, y); // area of interest

                    if ( ctrl_aoi == 1 && PAR_AFLAG == 1 ) { // average of impact probability:

                        Segment_get(&seg_impquant, &impquantold, x, y); // number of cases possibly affecting the pixel
                        if ( PAR_MULT == 0 ) Segment_get(&seg_relprobr, &relprobrold, x, y); // sum of release probability
                        Segment_get(&seg_impprobr, &impprobrold, x, y); // sum of impact probability
                        if ( impquantold > 0 ) {
                            if ( PAR_MULT == 0 ) relprobrnew = relprobrold / impquantold; // average zonal release probability
                            impprobrnew = impprobrold / impquantold; // average impact probability
                        }
                        else {
                            relprobrnew = 0; // average zonal release probability
                            impprobrnew = 0; // average impact probability
                        }
                        if ( PAR_MULT == 0 ) Segment_put(&seg_relprobr, &relprobrnew, x, y);
                        Segment_put(&seg_impprobr, &impprobrnew, x, y);
                    }
                    else if ( ctrl_aoi < 1 ) {

                        if ( PAR_MULT == 0 ) relprobrnew = sico.UDM;
                        impprobrnew = sico.UDM;
                        if ( PAR_MULT == 0 ) Segment_put(&seg_relprobr, &relprobrnew, x, y);
                        Segment_put(&seg_impprobr, &impprobrnew, x, y);
                    }

                #else

                    ctrl_aoi = seg_aoi[x][y]; // area of interest

                    if ( ctrl_aoi == 1 && PAR_AFLAG == 1 ) { // average of impact probability:

                        impquantold = seg_impquant[x][y]; // number of cases possibly affecting the pixel
                        if ( PAR_MULT == 0 ) relprobrold = seg_relprobr[x][y]; // sum of release probability
                        impprobrold = seg_impprobr[x][y]; // sum of impact probability
                        if ( impquantold > 0 ) {
                            if ( PAR_MULT == 0 ) relprobrnew = relprobrold / impquantold; // average zonal release probability
                            impprobrnew = impprobrold / impquantold; // average impact probability
                        }
                        else {
                            relprobrnew = 0; // average zonal release probability
                            impprobrnew = 0; // average impact probability
                        }
                        if ( PAR_MULT == 0 ) seg_relprobr[x][y] = relprobrnew;
                        seg_impprobr[x][y] =impprobrnew;
                    }
                    else if ( ctrl_aoi < 1 ) {

                        if ( PAR_MULT == 0 ) relprobrnew = sico.UDM;
                        impprobrnew = sico.UDM;
                        if ( PAR_MULT == 0 ) seg_relprobr[x][y] = relprobrnew;
                        seg_impprobr[x][y] = impprobrnew;
                    }

                #endif

            }
        }

        #ifdef WITHGRASS

            if ( PAR_MULT == 0 ) Segment_flush(&seg_relprobr);
            Segment_flush(&seg_impprobr);

        #endif
    }


// WRITING OUTPUT RASTER MAPS -----------------------------------------------------------------


    for ( model = 0; model < PAR_NMODELS; model++ ) {

        if ( PAR_MULT == 0 ) sprintf( mout, "%s_iM%d", PAR_PREFIX, PAR_MID[model] );
        else sprintf( mout, "%s_iM%d", PAR_PREFIX, xint );

        #ifdef WITHGRASS

            foutrasti( mout, seg_iM[model], sico );

        #else

            foutasci( seg_iM[model], PAR_MAPPATH, mout, sico );

        #endif
    }

    if ( PAR_MULT == 0 ) {

        sprintf( miis, "%s_iis", PAR_PREFIX );
        sprintf( msmov, "%s_smov", PAR_PREFIX );
        sprintf( mzmov, "%s_zmov", PAR_PREFIX );
        sprintf( momega, "%s_omega", PAR_PREFIX );
        sprintf( mimpquant, "%s_arz", PAR_PREFIX );

        if ( PAR_CTRLVEL == 1 ) sprintf( mxv, "%s_velocity", PAR_PREFIX );
        if ( PAR_CTRLIMPOBJ == 1 ) {

            sprintf( mris_impobj, "%s_iris", PAR_PREFIX );
            sprintf( msmovi, "%s_smovi", PAR_PREFIX );
            sprintf( mzmovi, "%s_zmovi", PAR_PREFIX );
            sprintf( momegai, "%s_omegai", PAR_PREFIX );
        }
        if ( PAR_PROB == 1 ) {
            sprintf( mrelprobr, "%s_prz", PAR_PREFIX );
            sprintf( mimpprobr, "%s_pi", PAR_PREFIX );
        }
        if ( PAR_SCORE == 1 ) sprintf( mhis, "%s_ihis", PAR_PREFIX );
    }
    else {

        sprintf( miis, "%s_iis%d", PAR_PREFIX, xint );
        if ( PAR_PROB == 1 ) sprintf( mimpprobr, "%s_pi%d", PAR_PREFIX, xint );
    }
    if ( PAR_CONN == 1 ) sprintf( mconn, "%s_sconn", PAR_PREFIX );

    #ifdef WITHGRASS

        foutrasti( miis, seg_iis, sico );

        if ( PAR_MULT == 0 ) {

            foutrastd( msmov, seg_smov, sico );
            foutrastd( mzmov, seg_zmov, sico );
            foutrastd( momega, seg_omega, sico );
            foutrasti( mimpquant, seg_impquant, sico );

            if ( PAR_CTRLVEL == 1 ) foutrastd( mxv, seg_xv, sico );
            if ( PAR_CTRLIMPOBJ == 1 ) {

                foutrasti( mris_impobj, seg_ris_impobj, sico );
                foutrastd( msmovi, seg_smovi, sico );
                foutrastd( mzmovi, seg_zmovi, sico );
                foutrastd( momegai, seg_omegai, sico );
            }
            if ( PAR_PROB == 1 ) foutrastd( mrelprobr, seg_relprobr, sico );
            if ( PAR_SCORE == 1 ) foutrasti( mhis, seg_his, sico );
        }
        if ( PAR_PROB == 1 ) foutrastd( mimpprobr, seg_impprobr, sico );
        if ( PAR_CONN == 1 ) foutrasti( mconn, seg_conn, sico );

    #else

        foutasci( seg_iis, PAR_MAPPATH, miis, sico );

        if ( PAR_MULT == 0 ) {

            foutascd( seg_smov, PAR_MAPPATH, msmov, sico );
            foutascd( seg_zmov, PAR_MAPPATH, mzmov, sico );
            foutascd( seg_omega, PAR_MAPPATH, momega, sico );
            foutasci( seg_impquant, PAR_MAPPATH, mimpquant, sico );

            if ( PAR_CTRLVEL == 1 ) foutascd( seg_xv, PAR_MAPPATH, mxv, sico );
            if ( PAR_CTRLIMPOBJ == 1 ) {

                foutasci( seg_ris_impobj, PAR_MAPPATH, mris_impobj, sico );
                foutascd( seg_smovi, PAR_MAPPATH, msmovi, sico );
                foutascd( seg_zmovi, PAR_MAPPATH, mzmovi, sico );
                foutascd( seg_omegai, PAR_MAPPATH, momegai, sico );
            }
            if ( PAR_PROB == 1 ) foutascd( seg_relprobr, PAR_MAPPATH, mrelprobr, sico );
            if ( PAR_SCORE == 1 ) foutasci( seg_his, PAR_MAPPATH, mhis, sico );
        }
        if ( PAR_PROB == 1 ) foutascd( seg_impprobr, PAR_MAPPATH, mimpprobr, sico );
        if ( PAR_CONN == 1 ) foutasci( seg_conn, PAR_MAPPATH, mconn, sico );

    #endif


// CLOSING FILES AND RELEASING MEMORY ---------------------------------------------------------


    fclose(f_summary);
    if ( PAR_MULT == 1 ) fclose(f_nout);
    if ( PAR_BACK == 1 ) fclose(f_backfile);

    if ( PAR_CONN == 1 && PAR_NCASES > 0 ) fclose(f_conn);

    #ifdef WITHGRASS

        Segment_release(&seg_aoi);
        Segment_release(&seg_elev);
        if ( PAR_CTRLFLOWDIR == 1 ) Segment_release(&seg_flowdir);
        if ( PAR_CTRLPROBMAP == 1 ) Segment_release(&seg_probmap);
        if ( PAR_CTRLRELEASEMAP == 1 ) Segment_release(&seg_releasemap);
        if ( PAR_X == 1 ) Segment_release(&seg_releasemap2);
        if ( PAR_CTRLSCOREMAP == 1 ) Segment_release(&seg_scoremap);
        if ( PAR_CTRLIMPOBJ == 1 ) Segment_release(&seg_impobj);
        if ( PAR_BACK == 1 ) Segment_release(&seg_imparea);

        Segment_release(&seg_iis);
        Segment_release(&seg_impquant);
        for ( model = 0; model < PAR_NMODELS; model++ ) Segment_release(&seg_iM[model]);

        if ( PAR_MULT == 0 ) {

            Segment_release(&seg_smov);
            Segment_release(&seg_zmov);
            Segment_release(&seg_omega);

            if ( PAR_CTRLVEL == 1 ) Segment_release(&seg_xv);
            if ( PAR_CTRLIMPOBJ == 1 ) {

                Segment_release(&seg_ris_impobj);
                Segment_release(&seg_smovi);
                Segment_release(&seg_zmovi);
                Segment_release(&seg_omegai);
            }
            if ( PAR_PROB == 1 ) {
                Segment_release(&seg_relprobr);
                Segment_release(&seg_impprob);
            }
            if ( PAR_SCORE == 1 ) Segment_release(&seg_his);

            if ( PAR_CONN == 1 ) Segment_release(&seg_conn);
        }
        if ( PAR_PROB == 1 ) Segment_release(&seg_impprobr);

    #else

        free(seg_aoi[0]); free(seg_aoi);
        free(seg_elev[0]); free(seg_elev);
        if ( PAR_CTRLFLOWDIR == 1 ) { free(seg_flowdir[0]); free(seg_flowdir); }
        if ( PAR_CTRLPROBMAP == 1 ) { free(seg_probmap[0]); free(seg_probmap); }
        if ( PAR_CTRLRELEASEMAP == 1 ) { free(seg_releasemap[0]); free(seg_releasemap); }
        if ( PAR_X == 1 ) { free(seg_releasemap2[0]); free(seg_releasemap2); }
        if ( PAR_CTRLSCOREMAP == 1 ) { free(seg_scoremap[0]); free(seg_scoremap); }
        if ( PAR_CTRLIMPOBJ == 1 ) { free(seg_impobj[0]); free(seg_impobj); }
        if ( PAR_BACK == 1 ) { free(seg_imparea[0]); free(seg_imparea); }

        free(seg_iis[0]); free(seg_iis);
        free(seg_impquant[0]); free(seg_impquant);
        free_imatrix3( seg_iM, PAR_NMODELS, sico.M );

        if ( PAR_MULT == 0 ) {

            free(seg_smov[0]); free(seg_smov);
            free(seg_zmov[0]); free(seg_zmov);
            free(seg_omega[0]); free(seg_omega);

            if ( PAR_CTRLVEL == 1 ) free(seg_xv[0]); free(seg_xv);
            if ( PAR_CTRLIMPOBJ == 1 ) {

                free(seg_ris_impobj[0]); free(seg_ris_impobj);
                free(seg_smovi[0]); free(seg_smovi);
                free(seg_zmovi[0]); free(seg_zmovi);
                free(seg_omegai[0]); free(seg_omegai);
            }
            if ( PAR_PROB == 1 ) {
                { free(seg_relprobr[0]); free(seg_relprobr); }
                { free(seg_impprob[0]); free(seg_impprob); }
            }
            if ( PAR_SCORE == 1 ) { free(seg_his[0]); free(seg_his); }
        }
        if ( PAR_PROB == 1 ) { free(seg_impprobr[0]); free(seg_impprobr); }

        free(PAR_MAPPATH);

    #endif

    free(sico.MAINMAPSET); free(PAR_PREFIX); free(PAR_FILEPATH); free(PAR_AOI); free(PAR_ELEV);
    if ( PAR_CTRLFLOWDIR == 1 ) free(PAR_FLOWDIR);
    if ( PAR_CTRLRELEASEMAP == 1 ) free(PAR_RELEASEMAP); if ( PAR_CTRLPROBMAP == 1 ) free(PAR_PROBMAP);
    if ( PAR_CTRLSCOREMAP == 1 ) free(PAR_SCOREMAP); if ( PAR_CTRLIMPOBJ == 1 ) free(PAR_IMPOBJ);
    if ( PAR_BACK == 1 ) free(PAR_IMPAREA); free(PAR_RUNUPMAX); free(PAR_FSLOPE); free(PAR_FDIR); free(PAR_SEG);
    free(PAR_SCTRL); free(PAR_MID); free(PAR_MTYPE); free(PAR_MPAR1); free(PAR_MPAR2); free(PAR_MPAR3);

    free(xadj); free(yadj); free(alphatestd); free(whtslope); free(xalpha); free(s_mov); free(omega); free(s_movmax);
    free(omega_stop); free(omega_test); free(xf_run); free(yf_run); free(ximpact); free(yimpact); free(iis); free(iM[0]); free(iM);
    free(s_mov_out); free(z_mov_out); free(omega_out); free(impquant);
    if ( PAR_CTRLVEL == 1 ) free(xxv); 
    //free(fdirtestd);

    if ( PAR_CTRLCDF == 1 ) { free(CDFO_O); free(CDFO); free(PDFO_O); free(PDFO); free(CDFL_O); free(CDFL); free(PDFL_O); free(PDFL); }
    if ( PAR_PROB == 1 ) { free(impprobr); free(impprob); }
    if ( PAR_SCORE == 1 ) free(his);
    if ( PAR_CTRLOBJSCORES == 1 ) { free(IMP_NAME); free(IMP_TYPE); free(IMP_S); free(ris_obj); }

    if ( PAR_NCASES > 0 ) {

        free(CASE_NAME); free(CASE_TYPE); free(CASE_V); free(CASE_Q); free(CASE_S); free(CASE_PROB);
	free(CASE_XR); free(CASE_YR); free(CASE_XS); free(CASE_YS);
    }


// EXIT ---------------------------------------------------------------------------------------


    printf("\n");
    exit(EXIT_SUCCESS);
}
