#!/usr/bin/env python3

##############################################################################
#
# MODULE:       r.randomwalk.mult.py
#
# AUTHOR:       Martin Mergili
# CONTRIBUTORS: Massimiliano Alvioli, Ivan Marchesini
#
# PURPOSE:      Conceptual tool for mass movement propagation
#               Script for multi-core processing
#
# COPYRIGHT:    (c) 2013 - 2021 by the author
#               (c) 2020 - 2021 by the University of Graz
#               (c) 2013 - 2021 by the BOKU University, Vienna
#               (c) 2015 - 2020 by the University of Vienna
#               (c) 2000 - 2021 by the GRASS Development Team
#
# VERSION:      20210525 (25 May 2021)
#
#               This program is free software under the GNU General Public
#               License (>=v2). Read the file COPYING that comes with GRASS
#               for details.
#
##############################################################################

#%module
#% description: Procedure for routing mass movements
#% keywords: Mass movement
#% keywords: Random walk
#% keywords: Raster
#% keywords: Routing
#%end

import grass.script as grass  # importing libraries
import os


def main():

    cellsize = os.environ["cellsize"]  # importing pixel size
    jid = os.environ["jid"]  # importing number of model run
    os.environ["XINT"] = jid  # exporting number of model run
    rtemp = os.environ["rtemp"]  # importing path to temporary directory
    pf = os.environ["pf"]  # importing path to temporary directory
    nflag2 = os.environ["nflag2"]  # importing control for randomized validation area

    grass.run_command("g.region", res=cellsize)  # setting pixel size

    print()
    print("Routing flow and exporting result.")
    print()

    grass.run_command(
        "r.randomwalk.main", flags="%s" % jid
    )  # executing r.randomwalk.main

    grass.run_command(
        "r.out.gdal",
        input=pf + "_iM" + str(jid),
        output=rtemp + "/" + pf + "_iM" + str(jid),
        format="GTiff",
        flags="c",
        quiet=True,
        overwrite=True,
    )  # exporting impact frequency map to tiff

    if nflag2 == "0":
        grass.run_command(
            "r.out.gdal",
            input=pf + "_iis" + str(jid),
            output=rtemp + "/" + pf + "_iis" + str(jid),
            format="GTiff",
            flags="c",
            quiet=True,
            overwrite=True,
        )  # exporting impact indicator score map to tiff

    elif nflag2 == "1":
        grass.run_command(
            "r.out.gdal",
            input=pf + "_pi" + str(jid),
            output=rtemp + "/" + pf + "_pi" + str(jid),
            format="GTiff",
            flags="c",
            quiet=True,
            overwrite=True,
        )  # exporting impact probability map to tiff

    print()
    print("Completed.")
    print()


if __name__ == "__main__":
    options, flags = grass.parser()
    main()
