#!/usr/bin/env python3

##############################################################################
#
# MODULE:       r.randomwalk.py
#
# AUTHOR:       Martin Mergili
# CONTRIBUTORS: Massimiliano Alvioli, Julia Krenn, Ivan Marchesini,
#               Markus Metz, Markus Neteler, Matthias Rauter
#
# PURPOSE:      Conceptual tool for mass movement propagation
#               Data management script
#
# COPYRIGHT:    (c) 2013 - 2021 by the author
#               (c) 2020 - 2021 by the University of Graz
#               (c) 2013 - 2021 by the BOKU University, Vienna
#               (c) 2015 - 2020 by the University of Vienna
#               (c) 2000 - 2021 by the GRASS Development Team
#
#               This program is free software under the GNU General Public
#               License (>=v2). Read the file COPYING that comes with GRASS
#               for details.
#
##############################################################################

#%module
#% description: Conceptual tool for mass movement propagation
#% keywords: Mass movement
#% keywords: Random walk
#% keywords: Raster
#% keywords: Routing
#%end

#%flag
#% key: a
#% description: Average of impact probability
#% guisection: flags
#%end

#%flag
#% key: b
#% description: Back-calculate observation
#% guisection: flags
#%end

#%flag
#% key: c
#% description: Compute connectivity
#% guisection: flags
#%end

#%flag
#% key: k
#% description: Keep result GRASS raster maps
#% guisection: flags
#%end

#%flag
#% key: m
#% description: Multiple model runs with random parameters
#% guisection: flags
#%end

#%flag
#% key: n
#% description: Multiple model runs with random release
#% guisection: flags
#%end

#%flag
#% key: p
#% description: Probabilities
#% guisection: flags
#%end

#%flag
#% key: q
#% description: Scores
#% guisection: flags
#%end

#%flag
#% key: s
#% description: Separate random walks for each model
#% guisection: flags
#%end

#%flag
#% key: v
#% description: Validation and visualization of results
#% guisection: flags
#%end

#%flag
#% key: x
#% description: Start routing from raster pixels
#% guisection: flags
#%end

#%option
#% key: prefix
#% type: string
#% description: Prefix for ouput files and folders
#% required: yes
#% multiple: no
#%end

#%option
#% key: cores
#% type: string
#% description: Number of cores to be used for multiple model runs (-m)
#% required: no
#% multiple: no
#%end

#%option
#% key: cellsize
#% type: string
#% description: Pixel size in metres
#% required: no
#% multiple: no
#%end

#%option
#% key: aoicoords
#% type: string
#% description: Set of coordinates delineating area of interest (N,S,E,W)
#% required: no
#% multiple: yes
#%end

#%option
#% key: aoimap
#% type: string
#% gisprompt: old,raster,cell
#% description: Name of area of interest raster map
#% required: no
#% multiple: no
#%end

#%option
#% key: elevation
#% type: string
#% gisprompt: old,raster,dcell
#% description: Name of elevation raster map
#% required: yes
#% multiple: no
#%end

#%option
#% key: flowdirection
#% type: string
#% gisprompt: old,raster,dcell
#% description: Name of flow direction raster map
#% required: no
#% multiple: no
#%end

#%option
#% key: releasefile
#% type: string
#% description: Path to text file with information on the cases to be routed
#% required: no
#% multiple: no
#%end

#%option
#% key: caserules
#% type: string
#% description: Rules for associating models to case types
#% required: no
#% multiple: no
#%end

#%option
#% key: releasemap
#% type: string
#% gisprompt: old,raster,cell
#% description: Name of raster map representing the ids of the cases to be routed
#% required: no
#% multiple: no
#%end

#%option
#% key: depositmap
#% type: string
#% gisprompt: old,raster,cell
#% description: Name of observed deposition areas raster map
#% required: no
#% multiple: no
#%end

#%option
#% key: impactmap
#% type: string
#% gisprompt: old,raster,cell
#% description: Name of observed impact areas raster map
#% required: no
#% multiple: no
#%end

#%option
#% key: probmap
#% type: string
#% gisprompt: old,raster,dcell
#% description: Name of raster map representing release probabilities of the cases to be routed
#% required: no
#% multiple: no
#%end

#%option
#% key: scoremap
#% type: string
#% gisprompt: old,raster,cell
#% description: Name of raster map representing release hazard indicator scores of the cases to be routed
#% required: no
#% multiple: no
#%end

#%option
#% key: impactobjects
#% type: string
#% gisprompt: old,raster,cell
#% description: Name of impact objects id raster map
#% required: no
#% multiple: no
#%end

#%option
#% key: objectscores
#% type: string
#% description: Path to file with scores of impact objects
#% required: no
#% multiple: no
#%end

#%option
#% key: models
#% type: string
#% description: Models to be used and related parameters
#% required: yes
#% multiple: yes
#%end

#%option
#% key: mparams
#% type: string
#% description: General parameters
#% required: yes
#% multiple: yes
#%end

#%option
#% key: sampling
#% type: string
#% description: Type of parameter sampling (positive number = random, 0 = controlled, negative number = OAT, -m)
#% required: no
#% multiple: no
#%end

#%option
#% key: retain
#% type: string
#% description: Per cent of cases to be retained for validation
#% required: no
#% multiple: no
#%end

#%option
#% key: functype
#% type: string
#% description: Type of probability density function of angle of reach
#% required: no
#% multiple: no
#%end

#%option
#% key: backfile
#% type: string
#% description: Path to back-calculation result file
#% required: no
#% multiple: no
#%end

#%option
#% key: cdffile
#% type: string
#% description: Path to cumulative probability densities for values of angle of reach
#% required: no
#% multiple: no
#%end

#%option
#% key: cdftype
#% type: string
#% description: Cdf to be used (1=angle of reach, 2=maximum travel distance)
#% required: no
#% multiple: no
#%end

#%option
#% key: zonalfile
#% type: string
#% description: Path to text file with correction and error parameters for zonal release probability
#% required: no
#% multiple: no
#%end

#%option
#% key: profile
#% type: string
#% description: Coordinates of profile along path of movement
#% required: no
#% multiple: yes
#%end


# Importing libraries

import grass.script as grass
import math
import os
import queue
import random
import shutil
import subprocess
import sys
import threading
import time


# Defining fundamental functions, classes and variables

queueLock = threading.Lock()  # queue for multi-core processing
workQueue = queue.Queue()

exitFlag = 0  # exit flag
ambvars = grass.gisenv()  # path to GRASS data
locpath = ambvars.GISDBASE + "/" + ambvars.LOCATION_NAME  # path to GRASS location
temppath = (
    ambvars.GISDBASE
    + "/"
    + ambvars.LOCATION_NAME
    + "/"
    + ambvars.MAPSET
    + "/.tmp/rtemp"
)  # path to temporary directory
scriptpath = os.environ["GISBASE"] + "/scripts"  # path to GRASS scripts
mainmapset = ambvars.MAPSET  # name of main mapset


def which(file):  # function for reading GRASS directory:
    for path in os.environ["PATH"].split(":"):
        if os.path.exists(path + "/" + file):
            return path + "/" + file
    return None


bingrass = which("grass70")  # reading GRASS directory


def ErrorMessage(errormessage):  # function for error message:
    grass.message(" ")
    grass.error(errormessage)
    grass.message(" ")
    sys.exit()


class myThread(threading.Thread):  # class for threading:
    def __init__(self, name, q):
        threading.Thread.__init__(self)
        self.name = name
        self.q = q

    def run(self):
        process_data(self.name, self.q)


def StartBatch(jid):  # function for start of multi-core processing:
    print("Executing model run %s" % jid)
    execute = "bash " + temppath + "/tmp" + str(jid) + "/batch" + str(jid)
    os.system(execute + " < /dev/null > " + temppath + "/out" + str(jid))
    print("Model run %s completed." % jid)
    return


def process_data(threadName, q):  # function supporting start of multi-core processing:
    while not exitFlag:
        queueLock.acquire()
        if not workQueue.empty():
            time.sleep(0.25)
            data = q.get()
            queueLock.release()
            StartBatch(data)
        else:
            queueLock.release()


def prepare_file(inFile, outFile):  # function for preparing input files:

    permfile = open(inFile, "r")  # opening file for reading
    permlines = sum(1 for line in permfile)  # counting number of lines
    permfile.close()  # closing file

    permfile = open(inFile, "r")  # again, opening file for reading
    tempfile = open(
        temppath + "/" + outFile + ".txt", "w"
    )  # opening temporary file for writing

    permfile.readline()  # reading header
    for permctr in range(1, permlines):  # for each line below header:
        permcont = permfile.readline()  # reading line
        permcont = permcont.replace("\n", "")
        print(
            permcont.replace("\t", "\n"), file=tempfile
        )  # putting everything into one column

    permfile.close()  # closing file
    tempfile.close()  # closing temporary file

    return permlines


def corrasc(corrname):  # function for correcting output ascii rasters

    os.rename(corrname + ".asc", corrname + "o.asc")

    fpressureo = open(corrname + "o.asc", "r")
    fpressure = open(corrname + ".asc", "w")

    fdata = ""
    for i in range(0, 5):
        fdatah = fpressureo.readline()
        fdata = fdata + "\t".join(fdatah.split()) + "\n"

    fdatah = fpressureo.readline()
    fdatai = fdatah.split()
    if fdatai[0] == "NODATA_value":
        fdata = fdata + fdatah
        fdataj = fdatai[1]
    else:
        fdata = fdata + "NODATA_value\t-9999\n" + fdatah
        fdataj = "-9999"

    fdata = fdata + "".join(fpressureo.readlines())
    fdata = fdata.replace("corner", "center").replace(fdataj, "-9999")

    fpressure.write(fdata)

    fpressureo.close()
    fpressure.close()

    os.system("rm -f " + corrname + "o.asc")
    os.system("rm -f " + corrname + ".asc.aux.xml")


def validate_results(
    pf, temppath, filepath, profile, depositmap, mflag, nruns, faimecp, cellsize
):
    # function for validation of results

    fval = open(
        filepath + pf + "_validation.txt", "w"
    )  # opening validation file for writing
    fvalstring = (
        "id\tArea\tomegat\tLmax"
    )  # adding header to string for validation file:

    fvalstring = fvalstring + "\t" + "Lim"
    if profile:
        if depositmap:
            fvalstring = fvalstring + "\t" + "Lobs" + "\t" + "Lratio"

    fvalstring = fvalstring + "\t" + "TP\tTN\tFP\tFN\tCSI\tHSS\tAUROC\tD2PC\tFoC\tSPI"
    if depositmap:
        grass.mapcalc(
            '"_obsbin"=if("%s">0,1,if("%s"==0,0,null()))' % (depositmap, depositmap),
            overwrite=True,
            quiet=True,
        )  # preparing observed impact area raster map

    print(fvalstring, file=fval)  # writing header to validation file

    fparamt = open(
        temppath + "/fparamt.txt", "r"
    )  # opening temporary input parameter file for reading
    fparami = open(
        filepath + pf + "_params.txt", "w"
    )  # opening input parameter file for reading
    print(fparamt.readline().replace("\n", ""), file=fparami)
    # writing header of temporary input parameter file to input parameter file

    nsuccess = 0  # intitalizing counter for number of successful simulations

    for jid in range(1, nruns + 1):  # loop over all model runs:

        if jid < 10:
            jfill = "00000" + str(jid)  # formatting model run string
        elif jid < 100:
            jfill = "0000" + str(jid)
        elif jid < 1000:
            jfill = "000" + str(jid)
        elif jid < 10000:
            jfill = "00" + str(jid)
        elif jid < 100000:
            jfill = "0" + str(jid)
        else:
            jfill = str(jid)

        if mflag:
            msuf = jfill  # suffix
        else:
            msuf = ""

        fnout = open(
            filepath + pf + "_parout" + str(jid) + ".txt", "r"
        )  # opening output file
        npixels = fnout.readline()  # reading number of affected pixels
        npixels = int(npixels.replace("\n", ""))  # removing newline
        smovmax = fnout.readline()  # reading maximum travel distance
        smovmax = float(smovmax.replace("\n", ""))  # removing newline
        omegamin = fnout.readline()  # reading angle of reach
        omegamin = float(omegamin.replace("\n", ""))  # removing newline
        csuccess = fnout.readline()  # reading control for success of simulation
        csuccess = int(csuccess.replace("\n", ""))  # removing newline
        fnout.close()  # closing file with number of time steps

        csuccess = 1
        if csuccess == 1:
            fvalstring = (
                jfill
                + "\t"
                + str(npixels * pow(cellsize, 2))
                + "\t"
                + str(omegamin)
                + "\t"
                + str(smovmax)
            )
        # writing number of model run, number of affected pixels, maximum travel distance and angle of reach to string

        flinet = fparamt.readline()  # reading line of temporary input parameter file

        if (
            csuccess == 1 and depositmap
        ):  # if simulation was successful and map of observed deposition area is provided:

            nsuccess = nsuccess + 1  # updating number of successful simulations

            if profile:  # if profile is defined:

                grass.run_command(
                    "r.profile",
                    input=depositmap,
                    output=temppath + "/fhrelease.txt",
                    coordinates=profile,
                    null_value="-9999",
                    overwrite=True,
                    quiet=True,
                )  # profile

                fprof = open(
                    temppath + "/fhrelease.txt", "r"
                )  # opening profile of release height for reading
                flines = sum(1 for line in fprof)  # counting number of lines
                fprof.close()  # closing file

                fprof = open(
                    temppath + "/fhrelease.txt"
                )  # again, opening profile file for reading
                for j in range(1, flines):  # loop over all lines:
                    fline = fprof.readline()  # reading and splitting line
                    fline = list(map(str, fline.split(" ")))
                    if not float(fline[2].replace("\n", "")) == 0:
                        flowline_start = float(
                            fline[1]
                        )  # identifying profile position at start of movement
                        break
                fprof.close()  # closing file

                grass.run_command(
                    "r.profile",
                    input=depositmap,
                    output=temppath + "/fimp.txt",
                    coordinates=profile,
                    null_value="-9999",
                    overwrite=True,
                    quiet=True,
                )  # profile of map

                flowline_stopimp = (
                    None
                )  # identifying profile position at terminus of observed movement:
                fprof = open(temppath + "/fimp.txt")  # opening profile for reading
                for j in range(1, flines):  # loop over all lines:
                    fline = fprof.readline()  # reading and splitting line
                    fline = list(map(str, fline.split(" ")))
                    if float(fline[2].replace("\n", "")) > 0 or (
                        not flowline_stopimp and j == flines - 1
                    ):
                        flowline_stopimp = float(
                            fline[1]
                        )  # identifying profile position at terminus of observed movement
                fprof.close()  # closing file
                maxlengthimp = flowline_stopimp - flowline_start  # movement length

                grass.run_command(
                    "r.profile",
                    input=pf + "_iis" + msuf,
                    output=temppath + "/f_iis.txt",
                    coordinates=profile,
                    null_value="-9999",
                    overwrite=True,
                    quiet=True,
                )  # profile of map

                flowline_stop = None
                fprof = open(temppath + "/f_iis.txt")  # opening profile for reading
                for j in range(1, flines):  # loop over all lines:
                    fline = fprof.readline()  # reading and splitting line
                    fline = list(map(str, fline.split(" ")))
                    if float(fline[2].replace("\n", "")) > 0 or (
                        not flowline_stop and j == flines - 1
                    ):
                        if not flowline_stop:
                            flowline_stop = flowline_start
                        else:
                            flowline_stop = float(
                                fline[1]
                            )  # identifying profile position at termination of movement
                fprof.close()  # closing file
                maxlength = flowline_stop - flowline_start  # movement length
                fvalstring = (
                    fvalstring + "\t" + str(maxlength)
                )  # adding movement length to string

                grass.run_command(
                    "r.profile",
                    input=depositmap,
                    output=temppath + "/fimp.txt",
                    coordinates=profile,
                    null_value="-9999",
                    overwrite=True,
                    quiet=True,
                )  # profile of map

                flowline_stopimp = (
                    None
                )  # identifying observed movement length and comparing to modelled one:
                fprof = open(temppath + "/fimp.txt")  # opening profile for reading
                for j in range(1, flines):  # loop over all lines:
                    fline = fprof.readline()  # reading and splitting line
                    fline = list(map(str, fline.split(" ")))
                    if float(fline[2].replace("\n", "")) > 0 or (
                        not flowline_stopimp and j == flines - 1
                    ):
                        flowline_stopimp = float(
                            fline[1]
                        )  # identifying profile position at terminus of observed movement
                fprof.close()  # closing file
                maxlengthimp = (
                    flowline_stopimp - flowline_start
                )  # observed movement length
                flowline_diffimp = flowline_stop - flowline_stopimp
                if maxlength > 0:
                    lratio = flowline_diffimp / maxlength
                else:
                    lratio = -9999
                fvalstring = fvalstring + "\t" + str(maxlengthimp) + "\t" + str(lratio)
                # adding observed movement length and difference in movement length to string

            else:
                fvalstring = fvalstring + "\t-9999\t-9999\t-9999"

            # Deriving TP, TN, FP and FN rates and AUCROC:

            grass.mapcalc(
                '"_simbin"=if("%s_iis%s">0,1,0)' % (pf, msuf),
                overwrite=True,
                quiet=True,
            )  # processed map of simulated deposit
            grass.mapcalc('"_eval"="_simbin"+10*"_obsbin"', overwrite=True, quiet=True)
            # combination of observed and simulated deposits
            grass.run_command(
                "r.stats",
                flags="c" + "n",
                separator="x",
                input="_eval",
                output="%s/validation.txt" % temppath,
                quiet=True,
                overwrite=True,
            )  # scores tp=11, tn=0, fp=1, fn=10

            valfile = open(
                temppath + "/validation.txt", "r"
            )  # opening raw file with validation scores
            valfile_lines = sum(1 for line in valfile)  # reading number of lines
            valfile.close()  # closing file

            if (
                valfile_lines < 4
            ):  # if number of lines is smaller than 4 (at least one of the scores is 0):
                valfile = open(temppath + "/validation.txt", "r")  # opening file again
                lr = valfile.readline()  # reading first line
                l1 = lr.find(
                    "0x"
                )  # checking for tn score (if not found, -1 is returned)
                if l1 > -1:
                    lr = valfile.readline()  # if found, reading next line
                l2 = lr.find(
                    "1x"
                )  # checking for fp score (if not found, -1 is returned)
                if l2 > -1:
                    lr = valfile.readline()  # if found, reading next line
                l3 = lr.find(
                    "10x"
                )  # checking for fn score (if not found, -1 is returned)
                if l3 > -1:
                    lr = valfile.readline()  # if found, reading next line
                l4 = lr.find(
                    "11x"
                )  # checking for tp score (if not found, -1 is returned)
                valfile.close()  # closing file again
            else:
                l1 = (
                    1
                )  # if number of lines is 4, setting all controls to positive (any value larger than -1 would be fine)
                l2 = 1
                l3 = 1
                l4 = 1

            valfile = open(
                temppath + "/validation.txt", "r"
            )  # once more opening temporary file with validation scores
            if l1 == -1:
                tn = float(0)  # if no tn score was identified before, setting to 0
            else:
                tn = valfile.readline()  # else, reading line
                tn = float(
                    tn.replace("0x", "").replace("\n", "")
                )  # extracting tn score
            if l2 == -1:
                fp = float(0)  # if no fp score was identified before, setting to 0
            else:
                fp = valfile.readline()  # else, reading line
                fp = float(
                    fp.replace("1x", "").replace("\n", "")
                )  # extracting fp score
            if l3 == -1:
                fn = float(0)  # if no fn score was identified before, setting to 0
            else:
                fn = valfile.readline()  # else, reading line
                fn = float(
                    fn.replace("10x", "").replace("\n", "")
                )  # extracting fn score
            if l4 == -1:
                tp = float(0)  # if no tp score was identified before, setting to 0
            else:
                tp = valfile.readline()  # else, reading line
                tp = float(
                    tp.replace("11x", "").replace("\n", "")
                )  # extracting tp score
            valfile.close()  # closing file for the last time

            tn = 10 * (tp + fn) - fp  # normalizing tn score
            if tp + fn > 0:
                foc = (tp + fp) / (tp + fn)  # conservativeness factor
            else:
                foc = -9999

            if tp + fn > 0 and tn + fp > 0:

                total = tp + tn + fp + fn  # sum of scores

                tpr = 100 * tp / total  # tp rate in percent
                tnr = 100 * tn / total  # tn rate in percent
                fpr = 100 * fp / total  # fp rate in percent
                fnr = 100 * fn / total  # fn rate in percent

                opr = tpr + fnr  # rate of positive observations
                onr = tnr + fpr  # rate of negative observations

                if not tpr + fpr + fnr == 0:
                    csi = tpr / (tpr + fpr + fnr)  # critical success index
                else:
                    csi = -9999
                if not (tpr + fnr) * (fnr + tnr) + (tpr + fpr) * (fpr + tnr) == 0:
                    hss = (2 * (tpr * tnr) - (fpr * fnr)) / (
                        (tpr + fnr) * (fnr + tnr) + (tpr + fpr) * (fpr + tnr)
                    )  # Heidke skill score
                else:
                    hss = -9999

                tprr = tpr / opr  # rate of true positive rate
                fprr = fpr / onr  # rate of false positive rate
                aucroc = (
                    (1 - fprr) * tprr + fprr * tprr / 2 + (1 - fprr) * (1 - tprr) / 2
                )  # area under the roc curve
                d2pc = pow(
                    pow(1 - tprr, 2) + pow(fprr, 2), 0.5
                )  # distance to perfect classification

            else:
                tpr = -9999
                tnr = -9999
                fpr = -9999
                fnr = -9999
                opr = -9999
                onr = -9999
                csi = -9999
                hss = -9999
                aucroc = -9999
                d2pc = -9999

            if foc > 0:
                logfoc = math.log10(foc)
            else:
                logfoc = -1000
            spi = (
                min(0.2, 0.2 * csi)
                + max(0, min(0.2, 0.2 * hss))
                + min(0.2, 0.2 * (1 - d2pc))
                + max(0, min(0.2, 0.2 * (1 - abs(2 * lratio))))
                + max(0, min(0.2, 0.2 * (1 - abs(logfoc))))
            )  # synthetic performance index

            fvalstring = (
                fvalstring
                + "\t"
                + str(tpr)
                + "\t"
                + str(tnr)
                + "\t"
                + str(fpr)
                + "\t"
                + str(fnr)
            )
            fvalstring = (
                fvalstring
                + "\t"
                + str(csi)
                + "\t"
                + str(hss)
                + "\t"
                + str(aucroc)
                + "\t"
                + str(d2pc)
                + "\t"
                + str(foc)
                + "\t"
                + str(spi)
            )
            # adding evaluation results to string

            print(fvalstring, file=fval)  # writing string for model run to text file

            flinep = flinet.replace("\t", ";").replace(
                "\n", ""
            )  # preparing line for writing to parameter input file for AIMEC
            print(
                flinep, file=faimecp
            )  # writing line of input parameter file to parameter input file for AIMEC

            grass.run_command(
                "r.out.gdal",
                input=pf + "_iis" + msuf,
                output=pf + "_results/" + pf + "_aimec/iis/" + jfill + ".asc",
                format="AAIGrid",
                quiet=True,
                overwrite=True,
            )  # exporting impact indicator score raster map to ascii
            grass.run_command(
                "r.out.gdal",
                input=depositmap,
                output=pf + "_results/" + pf + "_aimec/" + pf + "_depositarea.asc",
                format="AAIGrid",
                quiet=True,
                overwrite=True,
            )  # exporting deposit area raster map to ascii

            corrasc(pf + "_results/" + pf + "_aimec/iis/" + jfill)
            corrasc(pf + "_results/" + pf + "_aimec/" + pf + "_depositarea")

            flinei = flinet.replace(
                "\n", ""
            )  # preparing line for writing to input parameter file

        else:
            flinei = "*" + flinet.replace(
                "\n", ""
            )  # preparing line for writing to input parameter file
        print(
            flinei, file=fparami
        )  # writing line of temporary input parameter file to input parameter file

    fparamt.close()  # closing temporary input parameter file
    fparami.close()  # closing input parameter file
    fval.close()  # closing validation file
    if depositmap:
        faimecp.close()  # closing parameter input file for AIMEC


# Starting computation


def main():

    start = time.time()  # storing time

    # Reading flags and parameters

    aflag = flags["a"]
    bflag = flags["b"]
    cflag = flags["c"]
    kflag = flags["k"]
    mflag = flags["m"]
    nflag = flags["n"]
    pflag = flags["p"]
    qflag = flags["q"]
    sflag = flags["s"]
    vflag = flags["v"]
    xflag = flags["x"]

    pf = options["prefix"]
    mult = options["cores"]
    cellsize = options["cellsize"]
    aoicoords = options["aoicoords"]
    aoi = options["aoimap"]
    elevation = options["elevation"]
    flowdirection = options["flowdirection"]
    releasefile = options["releasefile"]
    caserules = options["caserules"]
    releasemap = options["releasemap"]
    depositmap = options["depositmap"]
    impactmap = options["impactmap"]
    probmap = options["probmap"]
    scoremap = options["scoremap"]
    impobj = options["impactobjects"]
    objscores = options["objectscores"]
    models = options["models"]
    mparams = options["mparams"]
    sampling = options["sampling"]
    retain = options["retain"]
    functype = options["functype"]
    cdffile = options["cdffile"]
    cdftype = options["cdftype"]
    backfile = options["backfile"]
    zonalfile = options["zonalfile"]
    profile = options["profile"]

    # Checking consistency of flag and parameter input

    if not pf:
        ErrorMessage("Please provide the parameter pf")
    if not elevation:
        ErrorMessage("Please provide the parameter elevation")
    if not models:
        ErrorMessage("Please provide the parameter models")
    if not mparams:
        ErrorMessage("Please provide the parameter mparams")  # mandatory parameters

    if aflag and not pflag:
        ErrorMessage(
            "The flag a may not be used together with the flag p"
        )  # inconsistent combinations of flags:
    if mflag and (nflag or pflag or qflag or sflag):
        ErrorMessage(
            "The flag m may not be used together with one of the flags n, p, q, or s"
        )
    if nflag and (mflag or qflag or sflag):
        ErrorMessage(
            "The flag n may not be used together with one of the flags m, q, or s"
        )
    if pflag and (mflag or bflag):
        ErrorMessage("The flag p may not be used together with one of the flags m or b")
    if qflag and (mflag or nflag or bflag):
        ErrorMessage("The flag q may not be used together with one of the flags n or b")
    if bflag and (nflag or pflag or qflag):
        ErrorMessage(
            "The flag b may not be used together with one of the flags n, p, or q"
        )

    if (mflag or nflag) and not sampling:
        ErrorMessage(
            "The parameter sampling has to be provided along with the flags m or n"
        )

    models = list(map(str, models.split(",")))  # splitting string with model parameters
    lenmodels = len(models)
    if not mflag and not sflag:
        lenmodel = float(5)  # correct length of list of model parameters:
    elif mflag and int(sampling) > 0:
        lenmodel = float(8)
    elif mflag:
        lenmodel = float(11)
    else:
        lenmodel = float(10)
    if not lenmodels % lenmodel == 0:
        ErrorMessage(
            "The number of comma-separated values provided for the parameter models is inappropriate"
        )
    nmodels = int(lenmodels / lenmodel)  # number of models

    ctrlvel = 0  # control for computation of velocity:
    if not mflag and not nflag:
        for i in range(0, nmodels, int(lenmodel)):
            if models[i + 1] == "5":
                ctrlvel = 1

    mparams = list(
        map(str, mparams.split(","))
    )  # splitting string with general parameters
    lenmparams = len(mparams)
    if not mflag and not sflag:
        lenmparam = float(7)  # correct length of list of general parameters:
    elif mflag and int(sampling) > 0:
        lenmparam = float(14)
    elif mflag:
        lenmparam = float(21)
    else:
        lenmparam = float(2)
    if not lenmparams % lenmparam == 0:
        ErrorMessage(
            "The number of comma-separated values provided for the parameter mparams is inappropriate"
        )

    if cdffile:
        cdflfile = cdffile.replace("_cdf", "_cdfl")
        cdfofile = cdffile.replace("_cdf", "_cdfo")
        pdflfile = cdffile.replace("_cdf", "_pdfl")
        pdfofile = cdffile.replace("_cdf", "_pdfo")
        if not cdftype:
            cdftype = 1

    if caserules:
        caserules = list(
            map(str, caserules.split(","))
        )  # splitting string with case rules
        lencaserules = len(caserules)
        if not lencaserules % float(nmodels + 1) == 0:
            ErrorMessage(
                "The number of comma-separated values provided for the parameter caserules is inappropriate"
            )
        # correct length of list of case rules
        ncasetypes = int(lencaserules / (nmodels + 1))  # number of case types

    if mflag and not mult:
        ErrorMessage(
            "The number of computational cores to be exploited (parameter cores) has to be provided along with the flag m"
        )
    if nflag and (not depositmap or not backfile):
        ErrorMessage(
            "The parameters depositmap and backfile have to be provided along with the flag n"
        )
    if bflag and not impactmap:
        ErrorMessage("The parameter impactmap has to be provided along with the flag b")
    if probmap and not pflag:
        ErrorMessage("The parameter probmap is only allowed with the flag p")
    if scoremap and not qflag:
        ErrorMessage("The parameter scoremap is only allowed with the flag q")
    if objscores and not impobj:
        ErrorMessage(
            "The parameter impobj has to be provided along with the parameter objscores"
        )

    if cflag and mflag:
        ErrorMessage("The flag c cannot be used together with the flag m")
    if cflag and not impobj:
        ErrorMessage("The flag c requires the parameter impobj")

    print()
    print()
    print(
        "##############################################################################"
    )
    print("#")
    print("# MODULE:       r.randomwalk 2.0 Pre-release")
    print("#")
    print("# AUTHOR:       Martin Mergili")
    print("# CONTRIBUTORS: Massimiliano Alvioli, Julia Krenn, Ivan Marchesini,")
    print("#               Markus Metz, Markus Neteler, Matthias Rauter")
    print("#")
    print("# PURPOSE:      The landslide routing tool")
    print("#")
    print("# COPYRIGHT:    (c) 2010 - 2021 by the author")
    print("#               (c) 2020 - 2021 by the University of Graz")
    print("#               (c) 2010 - 2021 by the BOKU University, Vienna")
    print("#               (c) 2015 - 2020 by the University of Vienna")
    print("#               (c) 2000 - 2021 by the GRASS Development Team")
    print("#               (C) 1993 - 2021 by the R Development Core Team")
    print("#")
    print("# VERSION:      20210531 (31 May 2021)")
    print("#")
    print("#")
    print("# PREFIX:       %s" % pf)
    print("#")
    print("#               This program is free software under the GNU General Public")
    print("#               License (>=v2). Read the file COPYING that comes with GRASS")
    print("#               for details.")
    print("#")
    print(
        "##############################################################################"
    )
    print()
    print()
    print("1. PREPARING ENVIRONMENT.")

    # Preparing environment

    grass.run_command(
        "g.remove", flags="f", type="rast", pattern=pf + "_*,_*", quiet=True
    )  # removing old temporary and result raster maps
    grass.run_command(
        "g.remove", flags="f", type="vect", pattern=pf + "v_*", quiet=True
    )  # removing old temporary vector maps

    os.system("rm -rf " + temppath)  # if necessary, deleting temporary directory
    os.system("mkdir " + temppath)  # creating temporary directory

    ascpath = pf + "_results/" + pf + "_ascii/"  # path to directory with ascii rasters
    filepath = (
        pf + "_results/" + pf + "_files/"
    )  # path to directory with result text files
    if vflag:
        plotpath = (
            pf + "_results/" + pf + "_plots/"
        )  # path to directory with result plots

    if 1 > 0:  # placeholder for possible condition to be implemented in the future

        if os.path.exists(pf + "_results"):
            os.system(
                "rm -rf " + pf + "_results"
            )  # if result directory already exists, deleting it
        os.mkdir(pf + "_results")  # creating directory for results
        os.mkdir(filepath)  # creating directory for result text files
        os.mkdir(ascpath)  # creating directory for result ascii rasters
        if vflag:
            os.mkdir(plotpath)  # creating directory for result plots

        # Defining GRASS region

        grass.run_command("g.region", flags="d")  # setting default region
        if cellsize:
            grass.run_command(
                "g.region", flags="a", res=cellsize
            )  # updating pixel size
        if aoicoords:  # updating extent:
            aoicoords = list(map(str, aoicoords.split(",")))
            grass.run_command(
                "g.region",
                flags="a",
                n=aoicoords[0],
                s=aoicoords[1],
                w=aoicoords[2],
                e=aoicoords[3],
            )

        if aoi:
            grass.mapcalc(
                '"_aoi"=if("%s">0,1,null())' % aoi, overwrite=True, quiet=True
            )  # area of interest
        else:
            grass.mapcalc('"_aoi"=pow("%s",0)' % elevation, overwrite=True, quiet=True)
        grass.mapcalc(
            '"_elevation"=int("%s"*"_aoi")' % elevation, overwrite=True, quiet=True
        )  # elevation raster map

        cellsize = grass.raster_info("_aoi")["nsres"]  # reading pixel size
        rnorth = grass.raster_info("_aoi")["north"]  # reading boundaries of aoi map
        rsouth = grass.raster_info("_aoi")["south"]
        rwest = grass.raster_info("_aoi")["west"]
        reast = grass.raster_info("_aoi")["east"]

        # Preprocessing area of interest map and input files

        if releasefile:
            releaselines = prepare_file(
                releasefile, "releasefile"
            )  # preparing release file
        else:
            releaselines = 1
        nreleases = releaselines - 1  # number of cases

        if cdffile:  # preparing cumulative and probability density function files
            prepare_file(cdfofile, "cdfo")
            prepare_file(cdflfile, "cdfl")
            prepare_file(pdfofile, "pdfo")
            prepare_file(pdflfile, "pdfl")

        if objscores:
            scorelines = prepare_file(
                objscores, "objscores"
            )  # preparing file with object scores
            nobjscores = scorelines - 1  # number of objects

        # Single model run

        if not mflag and not nflag:  # for single model run:

            # Writing model parameters to file

            pfile = open(temppath + "/param1.txt", "w")  # opening parameter file

            print(mainmapset, file=pfile)  # name of main mapset

            if aflag:
                print("1", file=pfile)  # average of impact probability
            else:
                print("0", file=pfile)
            if bflag:
                print("1", file=pfile)  # reconstruct observed impact area
            else:
                print("0", file=pfile)
            if cflag:
                print("1", file=pfile)  # compute connectivity
            else:
                print("0", file=pfile)
            print("0", file=pfile)  # single model run
            if pflag:
                print("1", file=pfile)  # probabilities
            else:
                print("0", file=pfile)
            if qflag:
                print("1", file=pfile)  # scores
            else:
                print("0", file=pfile)
            if sflag:
                print("1", file=pfile)  # separate random walks for each model
            else:
                print("0", file=pfile)
            if xflag:
                print("1", file=pfile)  # release from raster pixels
            else:
                print("0", file=pfile)

            print(pf, file=pfile)  # prefix
            print(filepath, file=pfile)  # path to output files
            print(ascpath, file=pfile)  # path to output ascii rasters

            print("_aoi", file=pfile)  # area of interest raster map
            print("_elevation", file=pfile)  # elevation raster map
            if flowdirection:
                print("1", file=pfile)
                print(flowdirection, file=pfile)  # flow direction raster map
            else:
                print("0", file=pfile)
            if probmap:
                print("1", file=pfile)
                print(probmap, file=pfile)  # release probability raster map
            else:
                print("0", file=pfile)
            if releasemap:
                print("1", file=pfile)
                print(releasemap, file=pfile)  # release areas raster map
            else:
                print("0", file=pfile)
            if scoremap:
                print("1", file=pfile)
                print(scoremap, file=pfile)  # release hazard indicator score raster map
            else:
                print("0", file=pfile)
            if impobj:
                print("1", file=pfile)
                print(impobj, file=pfile)  # raster map of impact objects
            else:
                print("0", file=pfile)
            if objscores:
                print("1", file=pfile)
                print(str(nobjscores), file=pfile)  # number of impact objects
            else:
                print("0", file=pfile)
            if bflag:
                print(impactmap, file=pfile)  # impact area raster map

            print(str(nreleases), file=pfile)  # number of cases
            print(str(nmodels), file=pfile)  # number of models
            print(str(ctrlvel), file=pfile)  # control for velocity

            for l in range(0, lenmparams):
                print(mparams[l], file=pfile)  # general parameters
            for l in range(0, lenmodels):
                print(models[l], file=pfile)  # model-specific parameters

            if caserules:
                print(str(ncasetypes), file=pfile)  # number of case types
                for l in range(0, lencaserules):
                    print(caserules[l], file=pfile)  # case rules
            else:
                print("0", file=pfile)

            if cdffile:
                print(str(cdftype), file=pfile)  # control for cdf and pdf files
            else:
                print("0", file=pfile)

            pfile.close()  # closing parameter file

            shutil.copyfile(
                temppath + "/param1.txt",
                pf + "_results/" + pf + "_files/" + pf + "_param.txt",
            )  # copying parameter file to results folder

            # Routing flow

            print("2. ROUTING MASS MOVEMENTS.")

            os.environ["XINT"] = "1"  # exporting id of model run
            grass.run_command(
                "r.randomwalk.main", flags="1"
            )  # executing r.randomwalk.main

            # Finalizing results

            print("3. FINALIZING RESULTS.")

            grass.mapcalc(
                '"%s_if"="_aoi"*0' % pf, overwrite=True, quiet=True
            )  # initializing impact frequency

            for model in range(1, nmodels + 1):
                grass.mapcalc(
                    '"%s_if"="%s_if"+"%s_iM%s"' % (pf, pf, pf, str(model)),
                    overwrite=True,
                    quiet=True,
                )  # updating impact frequency

            if pflag and probmap and not mparams[0] == "-9999":

                prepare_file(
                    zonalfile, "zparams"
                )  # preparing parameter file for zonal probability

                zpar = []
                zparfile = open(
                    temppath + "/zparams.txt", "r"
                )  # opening parameter file for reading:
                for zparams in range(0, 7):
                    zparfile.readline()
                    zpar.append(float(zparfile.readline().replace("\n", "")))
                zparfile.close()  # closing file

                grass.mapcalc(
                    '"%s_arz"="_aoi"*"%s_arz"*pow(%s,2)' % (pf, pf, cellsize),
                    overwrite=True,
                    quiet=True,
                )  # area of possible release zone
                grass.mapcalc(
                    '"%s_prz_stdev"=%s+log("%s_arz",10)*%s+pow(log("%s_arz",10),2)*%s+pow(log("%s_arz",10),3)*%s'
                    % (pf, zpar[3], pf, zpar[4], pf, zpar[5], pf, zpar[6]),
                    overwrite=True,
                    quiet=True,
                )  # standard deviation of zonal release probability
                grass.mapcalc(
                    '"%s_prz_stdev"=if("%s_prz_stdev"<0,0,"%s_prz_stdev")'
                    % (pf, pf, pf),
                    overwrite=True,
                    quiet=True,
                )  # finalizing standard deviation of zonal release probability
                grass.mapcalc(
                    '"%s_probcorr"=if("%s_arz"==0,"_aoi"*0,"_aoi"*(%s+(1-%s)*1/(1+exp(-(%s+log("%s_arz",10)*%s)))))'
                    % (pf, pf, zpar[0], zpar[0], zpar[1], pf, zpar[2]),
                    overwrite=True,
                    quiet=True,
                )  # average landslide probability for size of release area
                grass.mapcalc(
                    '"%s_prz"=if("%s_arz"==0,"_aoi"*0,round(1000*(1-((1-"%s")*(1-"%s_probcorr")/(1-%s))))/float(1000))'
                    % (pf, pf, probmap, pf, zpar[0]),
                    overwrite=True,
                    quiet=True,
                )  # zonal release probability
                grass.mapcalc(
                    '"%s_pic"=round(1000*"%s_prz"*"%s_pi")/float(1000)' % (pf, pf, pf),
                    overwrite=True,
                    quiet=True,
                )
                # composite impact probability
                grass.mapcalc(
                    '"%s_pi"=round(1000*"_aoi"*"%s_pi")/float(1000)' % (pf, pf),
                    overwrite=True,
                    quiet=True,
                )
                # rounding impact probability
                grass.mapcalc(
                    '"%s_arz"=round("_aoi"*"%s_arz")' % (pf, pf),
                    overwrite=True,
                    quiet=True,
                )  # rounding possible area of release zone
                grass.mapcalc(
                    '"%s_probcorr"=round(1000*"_aoi"*"%s_probcorr")/float(1000)'
                    % (pf, pf),
                    overwrite=True,
                    quiet=True,
                )
                # rounding correction function for zonal release probability

            elif pflag and probmap:

                maxif = grass.raster_info(pf + "_if")[
                    "max"
                ]  # maximum value of impact frequency
                grass.mapcalc(
                    '"%s_pic"=round((1000*float("%s_if")/float(%s)))/float(1000)'
                    % (pf, pf, maxif),
                    overwrite=True,
                    quiet=True,
                )
                # composite impact probability

        # Computing impact indicator scores or impact probability with randomized parameters and writing input files for AIMEC

        else:  # for multiple model runs:

            ncores = int(mult)  # number of processors to be used
            if nflag or int(sampling) > 0:
                nruns = int(sampling)  # number of model runs
            elif int(sampling) < 0:
                nruns = int(sampling) * -1

            if nflag:
                nflag2 = "1"  # control for validation with random subsets
            else:
                nflag2 = "0"

            # Preparing AIMEC input files

            if mflag:

                os.mkdir(
                    pf + "_results/" + pf + "_aimec"
                )  # creating directory for AIMEC input
                os.mkdir(
                    pf + "_results/" + pf + "_aimec/iis"
                )  # creating directory for impact maps for AIMEC input

                fparamt = open(
                    temppath + "/fparamt.txt", "w"
                )  # opening temporary input parameter file for writing
                print(
                    "id\tmid\tmtype\tmpar1\tmpar2\tmpar3\tnwalks\tlmin\tlctrl\tlseg\trmax\tfbeta\tfdir",
                    file=fparamt,
                )
                # header of temporary input parameter file

                if depositmap:

                    faimecp = open(
                        pf + "_results/" + pf + "_aimec/" + pf + "_aimec_params.txt",
                        "w",
                    )  # opening parameter file for AIMEC for writing
                    print(
                        "id, mid, mtype, mpar1, mpar2, mpar3, nwalks, lmin, lctrl, lseg, rmax, fbeta, fdir",
                        file=faimecp,
                    )
                    # header of parameter input file for AIMEC

                    if profile:  # creating profile file in directory for AIMEC input:

                        fpaimec = open(
                            pf + "_results/" + pf + "_aimec/" + pf + "_profile.xyz", "w"
                        )  # opening text file for writing
                        profcoord = list(
                            map(str, profile.split(","))
                        )  # splitting profile coordinate string
                        proflength = len(
                            profcoord
                        )  # reading number of profile coordinates
                        for profnum in range(
                            0, proflength, 2
                        ):  # loop over all coordinates:
                            profilex = profcoord[profnum]  # extracting x coordinate
                            profiley = profcoord[profnum + 1]  # extracting y coordinate
                            profstring = grass.read_command(
                                "r.what",
                                map="_elevation",
                                separator="\t",
                                coordinates=profilex + "," + profiley,
                                overwrite=True,
                                quiet=True,
                            )  # reading elevation value associated to profile coordinates
                            profstring = profstring.replace("\t\t", "\t").replace(
                                "\n", ""
                            )
                            print(
                                "%s" % profstring, file=fpaimec
                            )  # writing profile string to text file
                        fpaimec.close()  # closing file

                else:
                    faimecp = 0

            # Cumulative density functions of angles of reach for randomized subsets of the study areas

            if nflag:

                if retain:
                    retain2 = str(retain)
                else:
                    retain2 = "20"
                subprocess.call(
                    "Rscript %s/r.randomwalk.cdf.R %s %s %s %s %s %s %s %s"
                    % (
                        scriptpath,
                        temppath,
                        pf,
                        filepath,
                        backfile,
                        "2",
                        str(nruns),
                        str(retain2),
                        functype,
                    ),
                    shell=True,
                )

            # Preparing parameter sampling for multiple model runs

            if sampling == "0":  # if controlled sampling is applied:

                nruns0 = [1]
                nruns = 1

                for l in range(1, 4):
                    if float(models[3 * l - 1]) == float(models[3 * l]):
                        nruns0.append(1)
                    else:
                        nruns0.append(int(models[3 * l + 1]))
                    nruns *= nruns0[l]

                for l in range(0, 7):
                    if float(mparams[3 * l]) == float(mparams[3 * l + 1]):
                        nruns0.append(1)
                    else:
                        nruns0.append(int(mparams[3 * l + 2]))
                    nruns *= nruns0[l + 4]

                jids = [0]
                for l in range(1, 11):
                    jids.append(0)

            elif int(sampling) < 0:  # if OAT sampling is applied:

                nruns00 = nruns
                nruns = 0
                nvar = [0]

                for l in range(1, 4):
                    if not float(models[3 * l - 1]) == float(models[3 * l]):
                        nvar.append(1)
                        nruns += nruns00
                    else:
                        nvar.append(0)

                for l in range(0, 7):
                    if not float(mparams[3 * l]) == float(mparams[3 * l + 1]):
                        nvar.append(1)
                        nruns += nruns00
                    else:
                        nvar.append(0)

                lnrun = 0
                ltest = 1

            # Starting loop over all model runs

            for jid in range(1, nruns + 1):

                if jid < 10:
                    jfill = "00000" + str(jid)  # formatting model run string
                elif jid < 100:
                    jfill = "0000" + str(jid)
                elif jid < 1000:
                    jfill = "000" + str(jid)
                elif jid < 10000:
                    jfill = "00" + str(jid)
                elif jid < 100000:
                    jfill = "0" + str(jid)
                else:
                    jfill = str(jid)

                if sampling == "0":  # if controlled sampling is applied:

                    jidctrl = 0
                    for l in range(0, 11):

                        ll = l
                        if ll == 0 or jids[l] == 0 or jidctrl == 1:
                            jids[l] += 1
                        if jids[l] > nruns0[l]:
                            jids[l] = 1
                            jidctrl = 1
                        else:
                            jidctrl = 0

                if mflag and releasemap:
                    grass.mapcalc(
                        '"%s_releasemap%s"="%s"' % (pf, str(jid), releasemap),
                        overwrite=True,
                        quiet=True,
                    )
                # preparing map of observed release area

                # Preparing validation of cumulative density function of angles of reach

                if nflag:

                    prepare_file(
                        temppath + "/cdfox" + str(jid) + ".txt", "cdfo" + str(jid)
                    )  # preparing cdfo file for model run
                    prepare_file(
                        temppath + "/pdfox" + str(jid) + ".txt", "pdfo" + str(jid)
                    )  # preparing pdfo file for model run

                    prepare_file(
                        temppath + "/cdflx" + str(jid) + ".txt", "cdfl" + str(jid)
                    )  # preparing cdfl file for model run
                    prepare_file(
                        temppath + "/pdflx" + str(jid) + ".txt", "pdfl" + str(jid)
                    )  # preparing pdfl file for model run

                    cdfthrfile = open(
                        temppath + "/cdfthr" + str(jid) + ".txt", "r"
                    )  # reading randomization parameters  from files:
                    minsample = cdfthrfile.readline()  # minimum retained landslide id
                    minsample = minsample.replace("\n", "")
                    maxsample = cdfthrfile.readline()  # maximum retained landslide id
                    maxsample = maxsample.replace("\n", "")
                    cdfthrfile.close()  # closing file

                    if int(minsample) < int(maxsample):

                        if releasemap:
                            grass.mapcalc(
                                '"%s_releasemap%s"=if("%s">=%s&&"%s"<=%s,"%s",null())'
                                % (
                                    pf,
                                    str(jid),
                                    releasemap,
                                    minsample,
                                    releasemap,
                                    maxsample,
                                    releasemap,
                                ),
                                overwrite=True,
                                quiet=True,
                            )
                        # preparing map of observed release area

                        if depositmap:
                            grass.mapcalc(
                                '"_depositmap%s"=if("%s">=%s&&"%s"<=%s,"%s",0)'
                                % (
                                    str(jid),
                                    depositmap,
                                    minsample,
                                    depositmap,
                                    maxsample,
                                    depositmap,
                                ),
                                overwrite=True,
                                quiet=True,
                            )
                        # preparing map of observed deposition area

                    else:

                        if releasemap:
                            grass.mapcalc(
                                '"%s_releasemap%s"=if("%s">=%s||"%s"<=%s,"%s",null())'
                                % (
                                    pf,
                                    str(jid),
                                    releasemap,
                                    minsample,
                                    releasemap,
                                    maxsample,
                                    releasemap,
                                ),
                                overwrite=True,
                                quiet=True,
                            )
                        # preparing map of observed release area

                        if depositmap:
                            grass.mapcalc(
                                '"_depositmap%s"=if("%s">=%s||"%s"<=%s,"%s",0)'
                                % (
                                    str(jid),
                                    depositmap,
                                    minsample,
                                    depositmap,
                                    maxsample,
                                    depositmap,
                                ),
                                overwrite=True,
                                quiet=True,
                            )
                        # preparing map of observed deposition area

                # Writing parameters to file

                if mflag:

                    rmodels = [models[0]]  # randomizing model-specific parameters
                    rmodels.append(models[1])

                    if sampling == "0":  # if controlled sampling is applied:

                        ipar = []
                        for l in range(1, 4):
                            if float(models[3 * l - 1]) == float(models[3 * l]):
                                rmodels.append(float(models[3 * l - 1]))
                            else:
                                rmodels.append(
                                    float(models[3 * l - 1])
                                    + float(jids[l] - 1)
                                    / float(nruns0[l] - 1)
                                    * (float(models[3 * l]) - float(models[3 * l - 1]))
                                )
                                ipar.append(l)

                        if float(mparams[0]) == float(mparams[1]):
                            rmparams = [float(mparams[0])]
                        else:
                            rmparams = [
                                float(mparams[0])
                                + float(jids[4] - 1)
                                / float(nruns0[4] - 1)
                                * (float(mparams[1]) - float(mparams[0]))
                            ]
                        for l in range(1, 7):
                            if float(mparams[3 * l]) == float(mparams[3 * l + 1]):
                                rmparams.append(float(mparams[3 * l]))
                            else:
                                rmparams.append(
                                    float(mparams[3 * l])
                                    + float(jids[l + 4] - 1)
                                    / float(nruns0[l + 4] - 1)
                                    * (
                                        float(mparams[3 * l + 1])
                                        - float(mparams[3 * l])
                                    )
                                )
                                ipar.append(l + 3)

                    elif int(sampling) > 0:  # if random sampling is applied:

                        for l in range(1, 4):
                            rmodels.append(
                                random.uniform(
                                    float(models[2 * l]), float(models[2 * l + 1])
                                )
                            )

                        rmparams = [
                            random.uniform(float(mparams[0]), float(mparams[1]))
                        ]  # randomizing general parameters
                        for l in range(1, 7):
                            rmparams.append(
                                random.uniform(
                                    float(mparams[2 * l]), float(mparams[2 * l + 1])
                                )
                            )

                    else:  # if OAT sampling is applied:

                        if ltest < 11:
                            while nvar[ltest] == 0 and ltest < 11:
                                ltest += 1
                                lnrun = 0

                        for l in range(1, 4):
                            if (
                                float(models[3 * l - 1]) == float(models[3 * l])
                                or not ltest == l
                            ):
                                rmodels.append(float(models[3 * l + 1]))
                            else:
                                rmodels.append(
                                    float(models[3 * l - 1])
                                    + float(lnrun)
                                    / float(nruns00 - 1)
                                    * (float(models[3 * l]) - float(models[3 * l - 1]))
                                )

                        if float(mparams[0]) == float(mparams[1]) or not ltest == 4:
                            rmparams = [float(mparams[2])]
                        else:
                            rmparams = [
                                float(mparams[0])
                                + float(lnrun)
                                / float(nruns00 - 1)
                                * (float(mparams[1]) - float(mparams[0]))
                            ]
                        for l in range(1, 7):
                            if (
                                float(mparams[3 * l]) == float(mparams[3 * l + 1])
                                or not ltest == l + 4
                            ):
                                rmparams.append(float(mparams[3 * l + 2]))
                            else:
                                rmparams.append(
                                    float(mparams[3 * l])
                                    + float(lnrun)
                                    / float(nruns00 - 1)
                                    * (
                                        float(mparams[3 * l + 1])
                                        - float(mparams[3 * l])
                                    )
                                )

                        lnrun += 1
                        if lnrun == nruns00:
                            lnrun = 0
                            ltest += 1

                else:

                    rmodels = models  # applying parameters as given through input
                    rmparams = mparams

                pfile = open(
                    temppath + "/param" + str(jid) + ".txt", "w"
                )  # opening parameter file

                print(mainmapset, file=pfile)  # name of main mapset

                if aflag:
                    print("1", file=pfile)  # average of impact probability
                else:
                    print("0", file=pfile)
                if bflag:
                    print("1", file=pfile)  # reconstruct observed impact area
                else:
                    print("0", file=pfile)
                if cflag:
                    print("1", file=pfile)  # compute connectivity
                else:
                    print("0", file=pfile)
                print("1", file=pfile)  # multiple model runs
                if pflag:
                    print("1", file=pfile)  # probabilities
                else:
                    print("0", file=pfile)
                if qflag:
                    print("1", file=pfile)  # scores
                else:
                    print("0", file=pfile)
                print("0", file=pfile)  # separate random walks for each model disabled
                if xflag:
                    print("1", file=pfile)  # release from raster pixels
                else:
                    print("0", file=pfile)

                print(pf, file=pfile)  # prefix
                print(filepath, file=pfile)  # path to output files
                print(ascpath, file=pfile)  # path to output ascii rasters

                print("_aoi", file=pfile)  # area of interest raster map
                print("_elevation", file=pfile)  # elevation raster map
                if flowdirection:
                    print("1", file=pfile)
                    print(flowdirection, file=pfile)  # flow direction raster map
                else:
                    print("0", file=pfile)
                print("0", file=pfile)  # release probability raster map disabled
                if releasemap:
                    print("1", file=pfile)
                    print(
                        pf + "_releasemap" + str(jid), file=pfile
                    )  # release areas raster map
                else:
                    print("0", file=pfile)
                print(
                    "0", file=pfile
                )  # release hazard indicator score raster map disabled
                print("0", file=pfile)  # raster map of impact objects disabled
                print("0", file=pfile)  # impact objects disabled
                if bflag:
                    print(impactmap, file=pfile)  # impact area raster map

                print(str(nreleases), file=pfile)  # number of cases
                print("1", file=pfile)  # number of models restricted to 1
                print("0", file=pfile)  # no velocity to be computed

                for l in range(0, 7):
                    print(str(rmparams[l]), file=pfile)  # general parameters
                for l in range(0, 5):
                    print(str(rmodels[l]), file=pfile)  # model-specific parameters

                if caserules:
                    print(str(ncasetypes), file=pfile)  # number of case types
                    for l in range(0, lencaserules):
                        print(caserules[l], file=pfile)  # case rules
                else:
                    print("0", file=pfile)

                if nflag:
                    print("1", file=pfile)  # control for cdf and pdf files
                else:
                    print("0", file=pfile)

                pfile.close()  # closing parameter file

                if mflag:  # writing parameters to temporary input parameter file:

                    fparamtstring = jfill

                    for l in range(0, 5):
                        fparamtstring = fparamtstring + "\t" + str(rmodels[l])
                    for l in range(0, 7):
                        fparamtstring = fparamtstring + "\t" + str(rmparams[l])

                    print(
                        fparamtstring, file=fparamt
                    )  # writing line of temporary input parameter file to input parameter file

                # Creating batch file

                os.mkdir(temppath + "/tmp%s" % jid)  # creating directory for batch file
                strtmp = temppath + "/tmp%s/batch" % jid  # file name for batch file
                out = open(strtmp + str(jid), "w")  # creating batch file
                os.system(
                    "rm -rf " + locpath + "/map%s" % jid
                )  # removing old mapset for model run
                grass.run_command(
                    "g.mapset", flags="c", mapset="map%s" % jid
                )  # creating new mapset for model run
                grass.run_command(
                    "g.mapsets", mapset=ambvars.MAPSET, operation="add"
                )  # making original mapset active

                grass.run_command(
                    "g.mapset", mapset=ambvars.MAPSET
                )  # switching back to original mapset
                os.system(
                    "mkdir " + locpath + "/map%s/.tmp/rtemp" % jid
                )  # creating directory in mapset for model run

                print(
                    """#!/bin/bash
export jid=%s
export SHELL=\"/bin/bash\"
export cellsize=%s
export rtemp=%s
export pf=%s
export nflag2=%s
export GRASS_BATCH_JOB=%s/r.randomwalk.mult
rm -f ~/.grassrc6
cp %s/*.txt %s/map%s/.tmp/rtemp/
%s -text %s/map%s
unset GRASS_BATCH_JOB"""
                    % (
                        jid,
                        cellsize,
                        temppath,
                        pf,
                        nflag2,
                        scriptpath,
                        temppath,
                        locpath,
                        jid,
                        bingrass,
                        locpath,
                        jid,
                    ),
                    file=out,
                )

                out.close()  # closing batch file
                fd = os.open(strtmp + str(jid), os.O_RDONLY)  # opening batch file
                os.fchmod(fd, 0o755)  # making batch file executable
                os.close(fd)  # closing batch file

            if mflag:
                fparamt.close()  # closing input parameter file

            # Executing batch processing

            neff = min(ncores, nruns)
            threadList = list(range(1, neff + 1))
            nameList = list(range(1, nruns + 1))
            threads = []
            for tName in threadList:
                thread = myThread(tName, workQueue)
                thread.start()
                threads.append(thread)
            queueLock.acquire()
            for word in nameList:
                workQueue.put(word)
            queueLock.release()
            while not workQueue.empty():
                pass
            global exitFlag
            exitFlag = 1
            for t in threads:
                t.join()
            print()
            print("Batch processing completed. Collecting results.")
            print()

            # Finalizing result raster maps

            grass.mapcalc(
                '"%s_if"="_aoi"*0' % pf, overwrite=True, quiet=True
            )  # initializing impact frequency
            if mflag:
                grass.mapcalc(
                    '"%s_iis"="_aoi"*0' % pf, overwrite=True, quiet=True
                )  # initializing impact indicator score raster map

            if nflag:
                grass.mapcalc(
                    '"%s_picounter"="_aoi"*0' % pf, overwrite=True, quiet=True
                )  # initializing impact probability counter raster map
                grass.mapcalc(
                    '"%s_pi"="_aoi"*0' % pf, overwrite=True, quiet=True
                )  # initializing (average) impact probability raster map

            for jid in range(1, nruns + 1):  # loop over all model runs:

                if jid < 10:
                    jfill = "00000" + str(jid)  # formatting model run string
                elif jid < 100:
                    jfill = "0000" + str(jid)
                elif jid < 1000:
                    jfill = "000" + str(jid)
                elif jid < 10000:
                    jfill = "00" + str(jid)
                elif jid < 100000:
                    jfill = "0" + str(jid)
                else:
                    jfill = str(jid)

                os.system(
                    "rm -rf " + locpath + "/map" + str(jid)
                )  # removing mapsets for all model runs

                grass.run_command(
                    "r.in.gdal",
                    input=temppath + "/" + pf + "_iM" + str(jid),
                    output=pf + "_iM" + jfill,
                    overwrite=True,
                    quiet=True,
                )  # importing impact frequency raster map
                grass.mapcalc(
                    '"%s_if"="%s_if"+"%s_iM%s"' % (pf, pf, pf, jfill),
                    overwrite=True,
                    quiet=True,
                )  # updating impact frequency

                if mflag:
                    grass.run_command(
                        "r.in.gdal",
                        input=temppath + "/" + pf + "_iis" + str(jid),
                        output=pf + "_iis" + jfill,
                        overwrite=True,
                        quiet=True,
                    )  # importing impact indicator score raster map
                    grass.mapcalc(
                        '"%s_iis"=if("%s_iis%s">0,"%s_iis"+1,"%s_iis")'
                        % (pf, pf, jfill, pf, pf),
                        overwrite=True,
                        quiet=True,
                    )  # updating impact indicator score

                if nflag:
                    grass.run_command(
                        "r.in.gdal",
                        input=temppath + "/" + pf + "_pi" + str(jid),
                        output=pf + "_pi" + str(jid),
                        overwrite=True,
                        quiet=True,
                    )  # importing impact probability raster map

                    if aflag:
                        grass.mapcalc(
                            '"%s_pi"="%s_pi"+"%s_pi%s"' % (pf, pf, pf, str(jid)),
                            overwrite=True,
                            quiet=True,
                        )  # updating average impact probability
                    else:
                        grass.mapcalc(
                            '"%s_pi"=if("%s_pi%s">"%s_pi","%s_pi%s","%s_pi")'
                            % (pf, pf, str(jid), pf, pf, str(jid), pf),
                            overwrite=True,
                            quiet=True,
                        )  # updating impact probability

                    grass.mapcalc(
                        '"%s_picounter"=if("%s_pi%s">0,"%s_picounter"+1,"%s_picounter")'
                        % (pf, pf, str(jid), pf, pf),
                        overwrite=True,
                        quiet=True,
                    )  # updating impact probability counter

            if mflag:
                grass.mapcalc(
                    '"%s_iii"=float("%s_iis")/float(%s)' % (pf, pf, str(nruns)),
                    overwrite=True,
                    quiet=True,
                )  # impact indicator index

            if nflag and aflag:

                grass.mapcalc(
                    '"%s_pi"=if("%s_picounter"==0,0,round(1000*"%s_pi"/float("%s_picounter"))/float(1000))'
                    % (pf, pf, pf, pf),
                    overwrite=True,
                    quiet=True,
                )  # finalizing impact probability

                # Computing standard deviation of impact probability

                grass.mapcalc('"_pdiff"=0', overwrite=True, quiet=True)
                # initializing temporary raster map of squared difference between average probability and probability yielded by model run

                for jid in range(1, nruns + 1):  # loop over all model runs:

                    k = str(jid)

                    grass.mapcalc(
                        '"_pdiff"=if(isnull("%s_pi%s")==1||"%s_pi%s"==0,"_pdiff","_pdiff"+pow("%s_pi%s"-"%s_pi",2))'
                        % (pf, k, pf, k, pf, k, pf),
                        overwrite=True,
                        quiet=True,
                    )
                    # updating squared difference between average impact probability and impact probability yielded by model run

                grass.mapcalc(
                    '"%s_pi_stdev"=pow("_pdiff"/"%s_picounter",0.5)*"_aoi"' % (pf, pf),
                    overwrite=True,
                    quiet=True,
                )
                # raster map of standard deviation of impact probability

        stop = time.time()  # storing time (stop of main computation)
        comptime = stop - start  # computational time
        timefile = open(filepath + pf + "_time.txt", "w")
        timefile.write(str(comptime))  # writing computational time to file
        timefile.close()

        # Validating and visualizing results

        if vflag:

            print("4. VALIDATING AND VISUALIZING RESULTS.\n")

            if not nflag and not pflag:
                if not mflag:
                    nruns = 1
                else:
                    validate_results(
                        pf,
                        temppath,
                        filepath,
                        profile,
                        depositmap,
                        mflag,
                        nruns,
                        faimecp,
                        cellsize,
                    )

            grass.run_command(
                "r.relief",
                input="_elevation",
                output="_hillshade",
                quiet=True,
                overwrite=True,
            )  # temporary hillshade raster map
            grass.run_command(
                "r.out.gdal",
                flags="c",
                input="_hillshade",
                output=temppath + "/hillshade.asc",
                format="AAIGrid",
                quiet=True,
                overwrite=True,
            )  # exporing hillshade to ascii

            # Vector maps delineating release and impact areas

            if releasemap:  # if raster map of observed release areas is defined:

                releasedef = "1"
                grass.mapcalc(
                    '"_lreleasemap"=if("%s">0,1,null())' % releasemap,
                    overwrite=True,
                    quiet=True,
                )  # unary map of release areas
                grass.run_command(
                    "r.to.vect",
                    input="_lreleasemap",
                    output="v_lreleasemap",
                    type="area",
                    flags="v" + "s",
                    quiet=True,
                    overwrite=True,
                )  # line vector of observed release areas
                grass.run_command(
                    "v.out.ogr",
                    flags="c",
                    input="v_lreleasemap",
                    output=temppath,
                    format="ESRI_Shapefile",
                    type="boundary",
                    quiet=True,
                    overwrite=True,
                )  # exporting line vector of observed release areas to shape file

            else:
                releasedef = "0"  # control for availability of release area map

            if depositmap:  # if raster map of observed deposition areas is defined:

                depositdef = "1"
                grass.mapcalc(
                    '"_ldepositmap"=if("%s">0,1,null())' % depositmap,
                    overwrite=True,
                    quiet=True,
                )  # unary map of deposition areas
                grass.run_command(
                    "r.to.vect",
                    input="_ldepositmap",
                    output="v_ldepositmap",
                    type="area",
                    flags="v" + "s",
                    quiet=True,
                    overwrite=True,
                )  # line vector of observed deposition areas
                grass.run_command(
                    "v.out.ogr",
                    flags="c",
                    input="v_ldepositmap",
                    output=temppath,
                    format="ESRI_Shapefile",
                    type="boundary",
                    quiet=True,
                    overwrite=True,
                )  # exporting line vector of observed deposition areas to shape file

            else:
                depositdef = "0"  # control for availability of impact area raster map

            if impactmap:  # if raster map of observed impact areas is defined:

                impactdef = "1"
                grass.mapcalc(
                    '"_limpactmap"=if("%s">0,1,null())' % impactmap,
                    overwrite=True,
                    quiet=True,
                )  # unary map of impact areas
                grass.run_command(
                    "r.to.vect",
                    input="_limpactmap",
                    output="v_limpactmap",
                    type="area",
                    flags="v" + "s",
                    quiet=True,
                    overwrite=True,
                )  # line vector of observed impact areas
                grass.run_command(
                    "v.out.ogr",
                    flags="c",
                    input="v_limpactmap",
                    output=temppath,
                    format="ESRI_Shapefile",
                    type="boundary",
                    quiet=True,
                    overwrite=True,
                )  # exporting line vector of observed impact areas to shape file

            else:
                impactdef = "0"  # control for availability of impact area raster map

            if profile:  # if profile is defined, writing coordinates to file:

                profcoord = list(
                    map(str, profile.split(","))
                )  # splitting profile coordinate string
                proflength = len(profcoord)  # reading number of profile coordinates
                for profnum in range(0, proflength, 2):  # loop over all coordinates:
                    if profnum == 0:
                        profilex = profcoord[profnum]  # extracting x coordinate
                        profiley = profcoord[profnum + 1]  # extracting y coordinate
                    else:
                        profilex = (
                            profilex + "\t" + profcoord[profnum]
                        )  # extracting x coordinate
                        profiley = (
                            profiley + "\t" + profcoord[profnum + 1]
                        )  # extracting y coordinate

                fmapprof = open(
                    temppath + "/fmapprof.txt", "w"
                )  # opening text file for writing
                print(
                    "%s\n%s" % (profilex, profiley), file=fmapprof
                )  # writing profile string to text file
                fmapprof.close()  # closing file
                mapprof = "1"  # control for profile

            else:
                mapprof = "0"

            if releasefile and not xflag:
                releasefile2 = releasefile  # control for availability of release file
            else:
                releasefile2 = "0"

            if mflag:
                mflag2 = "1"  # control for multiple model runs (impact indicator score)
            else:
                mflag2 = "0"

            # Defining names of raster maps and output files for each map plot

            d_map = ["if"]  # mode=0
            if mflag:
                d_map.append("iii")  # mode=1
            else:
                d_map.append("iis")  # mode=1
            d_map.append("ihis")  # mode=2
            d_map.append("arz")  # mode=3
            d_map.append("prz")  # mode=4
            d_map.append("pi")  # mode=5
            d_map.append("pic")  # mode=6
            d_map.append("prz_stdev")  # mode=7
            d_map.append("pi_stdev")  # mode=8

            # Creating map plots

            for mode in range(0, 9):

                modectrl = (
                    0
                )  # defining validity of the different modes for different flags:
                if mode == 0:
                    modectrl = 1
                elif mflag and mode == 1:
                    modectrl = 1
                elif pflag and mode == 5:
                    modectrl = 1
                elif (
                    probmap
                    and not mparams[0] == "-9999"
                    and (mode == 3 or mode == 4 or mode == 6 or mode == 7)
                ):
                    modectrl = 1
                elif probmap and mparams[0] == "-9999" and mode == 6:
                    modectrl = 1
                elif qflag and mode <= 2:
                    modectrl = 1
                elif nflag and aflag and mode == 8:
                    modectrl = 1

                if modectrl == 1:

                    grass.run_command(
                        "r.out.gdal",
                        flags="c",
                        input=pf + "_" + d_map[mode],
                        output=temppath + "/" + d_map[mode] + ".asc",
                        format="AAIGrid",
                        quiet=True,
                        overwrite=True,
                    )  # exporting raster map to ascii

                    subprocess.call(
                        "Rscript %s/r.randomwalk.map.R %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s"
                        % (
                            scriptpath,
                            temppath,
                            "abc",
                            pf,
                            rnorth,
                            rsouth,
                            rwest,
                            reast,
                            cellsize,
                            d_map[mode],
                            "1",
                            "0",
                            "0",
                            "1",
                            str(mode),
                            "0",
                            "0",
                            "0",
                            impactdef,
                            releasedef,
                            releasefile2,
                            depositdef,
                            mflag2,
                            mapprof,
                        ),
                        shell=True,
                    )  # map with hillshade

            # Creating cumulative density plot of angles of reach

            if bflag and not mflag:
                subprocess.call(
                    "Rscript %s/r.randomwalk.cdf.R %s %s %s %s %s %s %s %s"
                    % (
                        scriptpath,
                        temppath,
                        pf,
                        filepath,
                        "abc",
                        "1",
                        "1",
                        "0",
                        functype,
                    ),
                    shell=True,
                )

            # Creating ROC plot

            if (pflag and ((not nflag and impactmap) or (nflag and depositmap))) or (
                mflag and depositmap
            ):

                if nflag:
                    mode = "2"
                    m = nruns - 1
                else:
                    mode = "1"
                    m = 0

                for mctr in range(
                    1, m + 2
                ):  # loop over all sets of probability and observation:

                    if (
                        nflag
                    ):  # random validation of impact probability (multiple model runs)
                        mstring = str(mctr)
                        obsstring = "_depositmap" + mstring
                        probstring = pf + "_pi" + mstring
                    else:  # composite impact probability
                        mstring = ""
                        if mflag:
                            obsstring = depositmap
                        else:
                            obsstring = impactmap
                        if mflag:
                            probstring = pf + "_iii"
                        elif probmap:
                            probstring = pf + "_pic"
                        else:
                            probstring = pf + "_pi"

                    grass.run_command(
                        "r.out.gdal",
                        flags="c",
                        input=obsstring,
                        output=temppath + "/observed" + mstring + ".asc",
                        format="AAIGrid",
                        quiet=True,
                        overwrite=True,
                    )  # exporting observation or iis map to ascii raster
                    grass.run_command(
                        "r.out.gdal",
                        flags="c",
                        input=probstring,
                        output=temppath + "/probability" + mstring + ".asc",
                        format="AAIGrid",
                        quiet=True,
                        overwrite=True,
                    )  # exporting temporary probability map to ascii raster

                subprocess.call(
                    "Rscript %s/r.randomwalk.roc.R %s %s %s %s %s --slave --quiet"
                    % (scriptpath, temppath, pf, mode, str(m), "1"),
                    shell=True,
                )
                # ROC plot relating observation to probability or iis (without normalization)

                subprocess.call(
                    "Rscript %s/r.randomwalk.roc.R %s %s %s %s %s --slave --quiet"
                    % (scriptpath, temppath, pf, mode, str(m), "2"),
                    shell=True,
                )
                # ROC plot relating observation to probability or iis (with normalization)

            # Producing graphics summarizing the validation results

            if mflag and depositmap:
                if sampling == "0" and len(ipar) == 2:

                    subprocess.call(
                        "Rscript %s/r.randomwalk.multval.R %s %s %s --slave --quiet"
                        % (scriptpath, pf, str(ipar[0]), str(ipar[1])),
                        shell=True,
                    )

    # Exporting result maps to ascii rasters

    grass.run_command(
        "r.out.gdal",
        flags="c",
        input=pf + "_if",
        output=ascpath + "/" + pf + "_if.asc",
        format="AAIGrid",
        quiet=True,
        overwrite=True,
    )  # impact frequency
    corrasc(ascpath + "/" + pf + "_if")

    if not mflag and not nflag:

        grass.run_command(
            "r.out.gdal",
            flags="c",
            input=pf + "_smov",
            output=ascpath + "/" + pf + "_smov.asc",
            format="AAIGrid",
            quiet=True,
            overwrite=True,
        )  # horizontal travel distance
        corrasc(ascpath + "/" + pf + "_smov")

        grass.run_command(
            "r.out.gdal",
            flags="c",
            input=pf + "_zmov",
            output=ascpath + "/" + pf + "_zmov.asc",
            format="AAIGrid",
            quiet=True,
            overwrite=True,
        )  # vertical travel distance
        corrasc(ascpath + "/" + pf + "_zmov")

        grass.run_command(
            "r.out.gdal",
            flags="c",
            input=pf + "_omega",
            output=ascpath + "/" + pf + "_omega.asc",
            format="AAIGrid",
            quiet=True,
            overwrite=True,
        )  # average angle of path
        corrasc(ascpath + "/" + pf + "_omega")

        if ctrlvel == 1:
            grass.run_command(
                "r.out.gdal",
                flags="c",
                input=pf + "_velocity",
                output=ascpath + "/" + pf + "_velocity.asc",
                format="AAIGrid",
                quiet=True,
                overwrite=True,
            )  # velocity
            corrasc(ascpath + "/" + pf + "_velocity")

    if pflag:
        grass.run_command(
            "r.out.gdal",
            flags="c",
            input=pf + "_pi",
            output=ascpath + "/" + pf + "_pi.asc",
            format="AAIGrid",
            quiet=True,
            overwrite=True,
        )  # impact probability
        corrasc(ascpath + "/" + pf + "_pi")

    if nflag:
        grass.run_command(
            "r.out.gdal",
            flags="c",
            input=pf + "_picounter",
            output=ascpath + "/" + pf + "_picounter.asc",
            format="AAIGrid",
            quiet=True,
            overwrite=True,
        )  # impact probability counter
        corrasc(ascpath + "/" + pf + "_picounter")

        if aflag:
            grass.run_command(
                "r.out.gdal",
                flags="c",
                input=pf + "_pi_stdev",
                output=ascpath + "/" + pf + "_pi_stdev.asc",
                format="AAIGrid",
                quiet=True,
                overwrite=True,
            )  # standard deviation of impact probability
            corrasc(ascpath + "/" + pf + "_pistdev")

    if probmap and not mparams[0] == "-9999":

        grass.run_command(
            "r.out.gdal",
            flags="c",
            input=pf + "_prz",
            output=ascpath + "/" + pf + "_prz.asc",
            format="AAIGrid",
            quiet=True,
            overwrite=True,
        )  # zonal release probability
        corrasc(ascpath + "/" + pf + "_prz")

        grass.run_command(
            "r.out.gdal",
            flags="c",
            input=pf + "_probcorr",
            output=ascpath + "/" + pf + "_probcorr.asc",
            format="AAIGrid",
            quiet=True,
            overwrite=True,
        )  # correction function for zonal release probability
        corrasc(ascpath + "/" + pf + "_probcorr")

        grass.run_command(
            "r.out.gdal",
            flags="c",
            input=pf + "_arz",
            output=ascpath + "/" + pf + "_arz.asc",
            format="AAIGrid",
            quiet=True,
            overwrite=True,
        )  # area of possible release zone
        corrasc(ascpath + "/" + pf + "_arz")

        if aflag:
            grass.run_command(
                "r.out.gdal",
                flags="c",
                input=pf + "_prz_stdev",
                output=ascpath + "/" + pf + "_prz_stdev.asc",
                format="AAIGrid",
                quiet=True,
                overwrite=True,
            )  # standard deviation of zonal release probability
            corrasc(ascpath + "/" + pf + "_prz_stdev")

    if probmap:

        grass.run_command(
            "r.out.gdal",
            flags="c",
            input=pf + "_pic",
            output=ascpath + "/" + pf + "_pic.asc",
            format="AAIGrid",
            quiet=True,
            overwrite=True,
        )  # composite impact probability
        corrasc(ascpath + "/" + pf + "_pic")

    if cflag:
        grass.run_command(
            "r.out.gdal",
            flags="c",
            input=pf + "_sconn",
            output=ascpath + "/" + pf + "_sconn.asc",
            format="AAIGrid",
            quiet=True,
            overwrite=True,
        )  # connectivity score
        corrasc(ascpath + "/" + pf + "_sconn")

    if mflag:
        grass.run_command(
            "r.out.gdal",
            flags="c",
            input=pf + "_iii",
            output=ascpath + "/" + pf + "_iii.asc",
            format="AAIGrid",
            quiet=True,
            overwrite=True,
        )  # impact indicator index
        corrasc(ascpath + "/" + pf + "_iii")

    if qflag:
        grass.run_command(
            "r.out.gdal",
            flags="c",
            input=pf + "_iis",
            output=ascpath + "/" + pf + "_iis.asc",
            format="AAIGrid",
            quiet=True,
            overwrite=True,
        )  # impact indicator score
        corrasc(ascpath + "/" + pf + "_iis")

        grass.run_command(
            "r.out.gdal",
            flags="c",
            input=pf + "_ihis",
            output=ascpath + "/" + pf + "_ihis.asc",
            format="AAIGrid",
            quiet=True,
            overwrite=True,
        )  # impact hazard indicator score
        corrasc(ascpath + "/" + pf + "_ihis")

        if impobj:
            grass.run_command(
                "r.out.gdal",
                flags="c",
                input=pf + "_iris",
                output=ascpath + "/" + pf + "_iris.asc",
                format="AAIGrid",
                quiet=True,
                overwrite=True,
            )  # impact risk indicator score
            corrasc(ascpath + "/" + pf + "_iris")

    # Cleaning file system and exiting

    grass.run_command(
        "g.remove", flags="f", type="rast", pattern="_*", quiet=True
    )  # removing temporary raster maps
    grass.run_command(
        "g.remove", flags="f", type="vect", pattern="v_*", quiet=True
    )  # removing temporary vector maps
    os.system("rm -rf " + temppath)  # removing temporary directory
    os.system(
        "rm -rf " + pf + "_results/" + pf + "_files/" + pf + "_parout*"
    )  # removing temporary output parameter files
    if (not pflag or not probmap) and not mflag:
        os.system("rm -rf " + pf + "_results/" + pf + "_files/" + pf + "_averages.txt")
    # removing output density and probability file

    if not kflag:
        grass.run_command(
            "g.remove", flags="f", type="rast", pattern=pf + "_*", quiet=True
        )  # removing result raster maps

    grass.run_command("g.region", flags="d")  # resetting default region

    print()
    print(
        "Completed in %.1f seconds (excluding validation and visualization)." % comptime
    )
    print("Please find the collected results in the directory %s_results." % pf)
    print()

    sys.exit()  # exit


if __name__ == "__main__":
    options, flags = grass.parser()
    main()
